/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata;

import static com.trifork.stamdata.Preconditions.checkNotNull;
import static java.lang.String.format;

import java.lang.annotation.Annotation;
import java.sql.SQLException;
import java.util.List;
import java.util.Map;

import org.hibernate.Query;
import org.hibernate.SQLQuery;
import org.hibernate.Session;
import org.hibernate.type.Type;
import org.joda.time.Instant;

import com.google.inject.Inject;
import com.trifork.stamdata.models.TemporalEntity;
import com.trifork.stamdata.persistence.Transactional;

import javax.persistence.Table;


public class Fetcher
{
    private final Session session;


    @Inject
    Fetcher(Session session)
    {
        this.session = checkNotNull(session);
    }
    
    @Transactional
    public <T extends TemporalEntity> T fetch(Class<T> type, Object id) throws SQLException
    {
        return fetch(Instant.now(), type, id);
    }


    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Class<T> type, String column, Object value, Type sqlType) throws SQLException
    {
        return fetch(Instant.now(), type, column, value, sqlType);
    }


    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Class<T> type, String column, Object value) throws SQLException
    {
        return fetch(Instant.now(), type, column, value);
    }


    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Class<T> type, Map<String, Object> columnAndValue) throws SQLException
    {
        return fetch(Instant.now(), type, columnAndValue);
    }


    @SuppressWarnings("unchecked")
    @Transactional
    public <T extends TemporalEntity> T fetch(Instant instant, Class<T> type, Object id) throws SQLException
    {
        checkNotNull(instant, "instant");
        checkNotNull(type, "type");
        checkNotNull(id, "id");

        // The database uses open/closed validity intervals: [a;b[
        // When a record is 'closed' (no longer valid) its ValidTo
        // is set to the same value as its ValidFrom. Therefore the
        // ValidTo must be checked using the '<' operator.

        String keyColumn = Entities.getIdColumnName(type);
        String entityName = type.getSimpleName();
        Annotation[] annotations = type.getAnnotations();
        for (Annotation annotation : annotations) {
            if (annotation instanceof Table) {
                Table t = (Table) annotation;
                entityName = t.name();
            }
        }

        // Ændret til SQL query for at kunne understøtte forskellige entiteter på samme tabel
        // Ikke optimalt men Hibernate er på vej ud uanset, og mysql bliver heller ikke lige udskiftet.
        String queryString = format("SELECT * FROM %s WHERE %s = :id AND ValidFrom <= :instant AND (:instant < ValidTo OR ValidTo IS NULL)", entityName, keyColumn);
        SQLQuery query = session.createSQLQuery(queryString);
        query.setParameter("id", id);
        query.setTimestamp("instant", instant.toDate());
        query.addEntity(type);
        // This query should only ever return a single result or the database's
        // structure has been corrupted. So we want an exception to be thrown.

        return (T) query.uniqueResult();
    }


    @SuppressWarnings("unchecked")
    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Instant instant, Class<T> type, String column, Object value, Type sqlType)
    {
        checkNotNull(instant, "instant");
        checkNotNull(type, "type");
        checkNotNull(column, "column");
        checkNotNull(value, "value");

        String entityName = type.getCanonicalName();
        Query query = session.createQuery(format("FROM %s WHERE %s = :value AND ValidFrom <= :instant AND (:instant < ValidTo OR ValidTo IS NULL)", entityName, column));

        query.setParameter("value", value, sqlType);
        query.setTimestamp("instant", instant.toDate());

        return (List<T>) query.list();
    }


    @SuppressWarnings("unchecked")
    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Instant instant, Class<T> type, String column, Object value)
    {
        checkNotNull(instant, "instant");
        checkNotNull(type, "type");
        checkNotNull(column, "column");
        checkNotNull(value, "value");

        String entityName = type.getCanonicalName();
        Query query = session.createQuery(format("FROM %s WHERE %s = :value AND ValidFrom <= :instant AND (:instant < ValidTo OR ValidTo IS NULL)", entityName, column));

        query.setParameter("value", value);
        query.setTimestamp("instant", instant.toDate());

        return (List<T>) query.list();
    }


    @SuppressWarnings("unchecked")
    @Transactional
    public <T extends TemporalEntity> List<T> fetch(Instant instant, Class<T> type, Map<String, Object> columnAndValue)
    {
        checkNotNull(instant, "instant");
        checkNotNull(type, "type");
        checkNotNull(columnAndValue, "columnAndValue");
        Preconditions.checkArgument(!columnAndValue.isEmpty(), "You must specify a map containing column name -> value pairs");

        StringBuffer whereClause = new StringBuffer();
        for (String columnName : columnAndValue.keySet())
        {
            if (whereClause.length() > 0)
            {
                whereClause.append(" AND ");
            }
            whereClause.append(columnName + " = :" + columnName);
        }
        whereClause.append(" AND ValidFrom <= :instant AND (:instant < ValidTo OR ValidTo IS NULL)");

        String entityName = type.getCanonicalName();
        Query query = session.createQuery(format("FROM %s WHERE " + whereClause.toString(), entityName));
        query.setTimestamp("instant", instant.toDate());

        // Set remaining variables on query
        for (String columnName : columnAndValue.keySet())
        {
            query.setParameter(columnName, columnAndValue.get(columnName));
        }

        return (List<T>) query.list();
    }
}
