/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata.persistence;

import static com.trifork.stamdata.Preconditions.checkArgument;
import static com.trifork.stamdata.Preconditions.checkNotNull;

import java.sql.Date;
import java.sql.Timestamp;

import com.trifork.stamdata.persistence.RecordSpecification.FieldSpecification;
import com.trifork.stamdata.persistence.RecordSpecification.RecordFieldType;

public class RecordBuilder
{
    private RecordSpecification recordSpecification;
    private Record record;

    public RecordBuilder(RecordSpecification recordSpecification)
    {
        this.recordSpecification = recordSpecification;
        record = new Record();
    }

    public RecordBuilder field(String fieldName, Object value)
    {
        if (value instanceof Long)
        {
            return field(fieldName, (Long)value);
        }
        else if (value instanceof String)
        {
            return field(fieldName, (String)value);
        }
        else
        {
            throw new IllegalArgumentException("Values in records must be string or long. field=" + fieldName);
        }
    }

    public RecordBuilder field(String fieldName, long value)
    {
        return field(fieldName, value, RecordSpecification.RecordFieldType.NUMERICAL);
    }

    public RecordBuilder field(String fieldName, String value)
    {
        return field(fieldName, value, RecordSpecification.RecordFieldType.ALPHANUMERICAL);
    }

    private RecordBuilder field(String fieldName, Object value, RecordSpecification.RecordFieldType recordFieldType)
    {
        checkNotNull(fieldName);
        checkArgument(value == null || getFieldType(fieldName) == recordFieldType, "Field " + fieldName + " is not " + recordFieldType);

        record = record.put(fieldName, value);

        return this;
    }

    public Record build()
    {
        if (recordSpecification.conformsToSpecifications(record))
        {
            return record;
        }
        else
        {
            throw new IllegalStateException("Mandatory fields not set");
        }
    }

    public Record addDummyFieldsAndBuild()
    {
        for(FieldSpecification fieldSpecification : recordSpecification.getFieldSpecs())
        {
            if(!record.containsKey(fieldSpecification.name))
            {
            if(fieldSpecification.type == RecordFieldType.ALPHANUMERICAL)
            {
                record = record.put(fieldSpecification.name, "D");
            }
            else if(fieldSpecification.type == RecordFieldType.TIMESTAMP)
            {
                record = record.put(fieldSpecification.name, Timestamp.valueOf("2000-01-01 00:00:00"));
            }
            else if(fieldSpecification.type == RecordSpecification.RecordFieldType.NUMERICAL)
            {
                record = record.put(fieldSpecification.name, 0l);
            }
            else
            {
                throw new AssertionError("");
            }
            }
        }
        return build();
    }

    private RecordSpecification.RecordFieldType getFieldType(String fieldName)
    {
        for (FieldSpecification fieldSpecification: recordSpecification.getFieldSpecs())
        {
            if (fieldSpecification.name.equals(fieldName))
            {
                return fieldSpecification.type;
            }
        }

        return null;
    }
}
