/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata.persistence;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.inject.Inject;

import org.joda.time.Instant;

import com.google.inject.Provider;


public class RecordFetcher
{
    private final Provider<Connection> connectionProvider;

    @Inject
    public RecordFetcher(Provider<Connection> connectionProvider)
    {
        this.connectionProvider = connectionProvider;
    }
    
    public Record fetchCurrent(String key, RecordSpecification recordSpecification,
                               String lookupColumn, Date currentTime) throws SQLException
    {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = String.format("SELECT * FROM %s WHERE %s = ? AND ValidFrom<=? AND (validTo IS NULL OR validTo>?)",
                    recordSpecification.getTable(), lookupColumn);
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, key);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                return createRecordFromResultSet(recordSpecification, resultSet);
            } else {
                return null;
            }
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }
    
    public Record fetchCurrent(String key, RecordSpecification recordSpecification, Date currentTime) throws SQLException
    {
        return fetchCurrent(key, recordSpecification, recordSpecification.getKeyColumn(), currentTime);
    }

    public List<RecordMetadata> fetchSince(RecordSpecification recordSpecification, long fromPID, Instant fromModifiedDate, int limit) throws SQLException
    {
        String queryString = String.format("SELECT * FROM %s WHERE " +
                "(PID > ? AND ModifiedDate = ?) OR " +
                "(ModifiedDate > ?) " +
                "ORDER BY ModifiedDate, PID LIMIT %d", recordSpecification.getTable(), limit);
        Connection connection = null;
        PreparedStatement preparedStatement = null;
        try {
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);

            Timestamp fromModifiedDateAsTimestamp = new Timestamp(fromModifiedDate.getMillis());
            preparedStatement.setObject(1, fromPID);
            preparedStatement.setTimestamp(2, fromModifiedDateAsTimestamp);
            preparedStatement.setTimestamp(3, fromModifiedDateAsTimestamp);

            ResultSet resultSet = preparedStatement.executeQuery();

            List<RecordMetadata> result = new ArrayList<RecordMetadata>();
            while(resultSet.next())
            {
                Instant validFrom = new Instant(resultSet.getTimestamp("ValidFrom"));
                // Validto can be null - and a new Instant with null as argument gives "now"
                Timestamp vto = resultSet.getTimestamp("ValidTo");
                Instant validTo = null;
                if(vto != null) {
                    validTo = new Instant(vto);
                }
                Instant modifiedDate = new Instant(resultSet.getTimestamp("ModifiedDate"));
                Long pid = (Long) resultSet.getObject("PID");
                Record record = createRecordFromResultSet(recordSpecification, resultSet);
                RecordMetadata recordMetadata = new RecordMetadata(validFrom, validTo, modifiedDate, pid, record);
                result.add(recordMetadata);
            }
            return result;
        } finally {
            if (preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    private Record createRecordFromResultSet(RecordSpecification recordSpecification, ResultSet resultSet) throws SQLException
    {
        RecordBuilder builder = new RecordBuilder(recordSpecification);

        for (RecordSpecification.FieldSpecification fieldSpec : recordSpecification.getFieldSpecs())
        {
            if(fieldSpec.persistField)
            {
                String fieldName = fieldSpec.name;

                if (fieldSpec.type == RecordSpecification.RecordFieldType.NUMERICAL)
                {
                    builder.field(fieldName, resultSet.getLong(fieldName));
                }
                else if (fieldSpec.type == RecordSpecification.RecordFieldType.ALPHANUMERICAL)
                {
                    builder.field(fieldName, resultSet.getString(fieldName));
                }
                else
                {
                    throw new AssertionError("Invalid field specifier used");
                }
            }
        }

        return builder.build();
    }
}
