/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata.persistence;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.List;

import com.google.inject.Inject;
import com.google.inject.Provider;
import com.trifork.stamdata.Preconditions;
import org.apache.commons.lang.StringUtils;
import org.joda.time.Instant;

import com.google.common.collect.Lists;
import com.trifork.stamdata.persistence.RecordSpecification.FieldSpecification;

public class RecordPersister
{
    private final Provider<Connection> connectionProvider;
    private final Instant transactionTime;

    @Inject
    public RecordPersister(Provider<Connection> connectionProvider, Instant transactionTime)
    {
        this.connectionProvider = connectionProvider;
        this.transactionTime = transactionTime;
    }

    public void persist(Record record, RecordSpecification specification) throws SQLException
    {
        Preconditions.checkNotNull(record);
        Preconditions.checkNotNull(specification);
        Preconditions.checkArgument(specification.conformsToSpecifications(record));

        Connection connection = null;
        PreparedStatement insertRecordStatement = null;
        try {
            connection = connectionProvider.get();
            insertRecordStatement = connection.prepareStatement(createInsertStatementSql(specification));

            populateInsertStatement(insertRecordStatement, record, specification);
            insertRecordStatement.execute();
        } finally {
            if (insertRecordStatement != null) {
                insertRecordStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    public void populateInsertStatement(PreparedStatement preparedStatement, Record record,
                                        RecordSpecification recordSpec) throws SQLException
    {
        Preconditions.checkArgument(recordSpec.conformsToSpecifications(record), "The record does not conform to it's spec.");

        int index = 1;

        for (FieldSpecification fieldSpecification: recordSpec.getFieldSpecs())
        {
            if(fieldSpecification.persistField)
            {
                if (fieldSpecification.type == RecordSpecification.RecordFieldType.ALPHANUMERICAL)
                {
                    preparedStatement.setString(index, (String) record.get(fieldSpecification.name));
                }
                else if (fieldSpecification.type == RecordSpecification.RecordFieldType.NUMERICAL)
                {
                    preparedStatement.setLong(index, (Long) record.get(fieldSpecification.name));
                }
                else if (fieldSpecification.type == RecordSpecification.RecordFieldType.TIMESTAMP)
                {
                    preparedStatement.setTimestamp(index, (Timestamp) record.get(fieldSpecification.name));
                }
                else
                {
                    throw new AssertionError("RecordType was not set correctly in the specification");
                }

                index++;
            }
        }

        preparedStatement.setTimestamp(index++, new Timestamp(transactionTime.getMillis()));
        preparedStatement.setTimestamp(index++, new Timestamp(transactionTime.getMillis()));
    }

    public String createInsertStatementSql(RecordSpecification specification)
    {
        StringBuilder builder = new StringBuilder();

        builder.append("INSERT INTO ").append(specification.getTable()).append(" (");

        List<String> fieldNames = Lists.newArrayList();
        List<String> questionMarks = Lists.newArrayList();

        for (FieldSpecification fieldSpecification: specification.getFieldSpecs())
        {
            if(fieldSpecification.persistField)
            {
                fieldNames.add(fieldSpecification.name);
                questionMarks.add("?");
            }
        }

        fieldNames.add("ValidFrom");
        questionMarks.add("?");

        fieldNames.add("ModifiedDate");
        questionMarks.add("?");

        builder.append(StringUtils.join(fieldNames, ", "));
        builder.append(") VALUES (");
        builder.append(StringUtils.join(questionMarks, ", "));
        builder.append(")");

        return builder.toString();
    }
}
