/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package com.trifork.stamdata.persistence;

import java.util.ArrayList;
import java.util.Arrays;
import java.sql.Timestamp;
import java.util.List;

import com.google.common.collect.ImmutableList;
import com.trifork.stamdata.Preconditions;

/**
 * @author Thomas G. Kristensen <tgk@trifork.com>
 */
public class RecordSpecification
{
    private final String table;
    private final String keyColumn;

    public String getTable()
    {
        return table;
    }

    public String getKeyColumn()
    {
        return keyColumn;
    }

    public static enum RecordFieldType
    {
        ALPHANUMERICAL,
        NUMERICAL,
        TIMESTAMP
    }

    public static enum Modifiers
    {
        DO_NOT_PERSIST;
    }

    public static class FieldSpecification
    {
        public final String name;
        public final RecordFieldType type;
        public final int length;
        public final boolean persistField;

        public FieldSpecification(String name, RecordFieldType type, int length, boolean persistField)
        {
            this.name = name;
            this.type = type;
            this.length = length;
            this.persistField = persistField;
        }

        /**
         * Returns a copy of the field with a numerical entry
         */
        public FieldSpecification numerical()
        {
            return new FieldSpecification(name, RecordFieldType.NUMERICAL, length, persistField);
        }

        /**
         * Returns a copy of the field with a numerical entry
         */
        public FieldSpecification date()
        {
            return new FieldSpecification(name, RecordFieldType.TIMESTAMP, length, persistField);
        }

        /**
         * Returns a copy of the field that will not be persisted
         */
        public FieldSpecification doNotPersist()
        {
            return new FieldSpecification(name, type, length, false);
        }
    }

    /**
     * Creates an alphanumerical field that will be persisted
     */
    public static FieldSpecification field(String name, int length)
    {
        return new FieldSpecification(name, RecordFieldType.ALPHANUMERICAL, length, true);
    }

    private List<FieldSpecification> fields;

    private RecordSpecification(String table, String keyColumn)
    {
        this.table = table;
        this.keyColumn = keyColumn;

        fields = new ArrayList<FieldSpecification>();
    }

    public static RecordSpecification createSpecification(String tableName, String keyColumnName, FieldSpecification... fieldSpecifications)
    {
        RecordSpecification recordSpecification = new RecordSpecification(tableName, keyColumnName);

        recordSpecification.fields = Arrays.asList(fieldSpecifications);

        return recordSpecification;
    }

    public Iterable<FieldSpecification> getFieldSpecs()
    {
        return ImmutableList.copyOf(fields);
    }

    public int acceptedTotalLineLength()
    {
        int totalLength = 0;
        for(FieldSpecification fieldSpecification: fields)
        {
            totalLength += fieldSpecification.length;
        }
        return totalLength;
    }

    public boolean conformsToSpecifications(Record record)
    {
        Preconditions.checkNotNull(record, "record");

        for (FieldSpecification fieldsSpecification: fields)
        {
            if(fieldsSpecification.persistField)
            {
                if (!record.containsKey(fieldsSpecification.name))
                {
                    return false;
                }
                else
                {
                    Object value = record.get(fieldsSpecification.name);

                    if (fieldsSpecification.type == RecordFieldType.NUMERICAL)
                    {
                        if ((value instanceof Long) || (value instanceof Integer)) {
                            continue;
                        } else {
                        	return false;
                        }
                    }
                    else if (fieldsSpecification.type == RecordFieldType.TIMESTAMP)
                    {
                        if (value instanceof Timestamp) {
                            continue;
                        } else {
                            return false;
                        }
                    }
                    else if (fieldsSpecification.type == RecordFieldType.ALPHANUMERICAL)
                    {
                        if (value != null && !(value instanceof String))
                        {
                            return false;
                        }
                        else if (value != null)
                        {
                            String valueAsString = String.valueOf(value);

                            if (valueAsString.length() > fieldsSpecification.length)
                            {
                                return false;
                            }
                        }
                    }
                    else
                    {
                        throw new AssertionError("Field specification is in illegal state. Type must be set.");
                    }
                }
            }
        }

        return true;
    }
}
