/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.security;

import org.apache.log4j.Logger;
import org.hibernate.Session;
import org.hibernate.SessionFactory;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

/**
 * In order to use the same connection pooling mechanism as the Hibernate queries, this class
 * uses the workaround of being passed a Hibernate Session in its constructor and using
 * sessionFactory.doWork(...) to execute it's native SQL queries
 */
public class WhitelistServiceImpl implements WhitelistService {
	private static final Logger logger = Logger.getLogger(WhitelistServiceImpl.class);
	private static final String QUERY_IS_CVR_WHITELISTED = "SELECT cvr FROM whitelist_config WHERE component_name=? AND cvr=?";
	private static final String QUERY_GET_ALL_WHITELISTED = "SELECT cvr FROM whitelist_config WHERE component_name=?";

	private SessionFactory sessionFactory;

	public WhitelistServiceImpl(SessionFactory sessionFactory) {
		this.sessionFactory = sessionFactory;
	}

	private PreparedStatement prepareStatement(Connection con, String query) {
		try {
			con.setReadOnly(true);
		} catch (SQLException e) {
			throw new RuntimeException("Unable to set Read Only on connection", e);
		}

		PreparedStatement statement;
		try {
			statement = con.prepareStatement(query);
		} catch (SQLException e) {
			throw new RuntimeException("Unable to prepare statement for query " + query + " exception="+e.getMessage(), e);
		}

		return statement;
	}

	@Override
	public List<String> getWhitelist(final String serviceName) {
		final List<String> cvrs = new ArrayList<String>();
        Session session = null;
        try {
            session = sessionFactory.openSession();
            session.beginTransaction();
            Connection con = session.connection();
			PreparedStatement getAllWhitelisted = prepareStatement(con, QUERY_GET_ALL_WHITELISTED);
			getAllWhitelisted.setString(1, serviceName);
			ResultSet resultSet = getAllWhitelisted.executeQuery();
			while (resultSet.next()) {
				cvrs.add(resultSet.getString("cvr"));
			}
			resultSet.close();
		} catch (SQLException e) {
			throw new RuntimeException("Unable to get whitelisted CVR numbers using prepared statement", e);
		} finally {
			close(session);
		}
		return cvrs;
	}

	@Override
	public boolean isCvrWhitelisted(final String cvr, final String serviceName) {
		final List<String> cvrs = new ArrayList<String>();
		Session session = null;
		try {
            session = sessionFactory.openSession();
            session.beginTransaction();
            Connection con = session.connection();
			PreparedStatement isCVRWhitelisted = prepareStatement(con, QUERY_IS_CVR_WHITELISTED);
			isCVRWhitelisted.setString(1, serviceName);
			isCVRWhitelisted.setString(2, cvr);
			ResultSet resultSet = isCVRWhitelisted.executeQuery();
			while (resultSet.next()) {
				cvrs.add(resultSet.getString("cvr"));
			}
			resultSet.close();
		} catch (SQLException e) {
			throw new RuntimeException("Unable to check if CVR is whitelisted using prepared statement", e);
		} finally {
			close(session);
		}
		return cvrs.size() > 0;
	}

	private void close(Session session) {
		if (session != null) {
			session.close();
		}
	}
}
