/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.replication.webservice;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.math.BigInteger;
import java.net.URL;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.List;
import java.util.Scanner;
import java.util.Properties;

import javax.servlet.FilterChain;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.ws.Holder;
import javax.xml.ws.soap.SOAPFaultException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import com.google.inject.Provider;

import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.BOMInputStream;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.joda.time.DateTime;
import org.joda.time.Instant;
import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mortbay.jetty.Request;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXParseException;

import com.google.common.collect.Lists;
import com.google.inject.Inject;
import com.trifork.stamdata.jaxws.SealNamespaceResolver;
import com.trifork.stamdata.persistence.Record;
import com.trifork.stamdata.persistence.RecordBuilder;
import com.trifork.stamdata.persistence.RecordMySQLTableGenerator;
import com.trifork.stamdata.persistence.RecordPersister;
import com.trifork.stamdata.persistence.RecordSpecification;
import com.trifork.stamdata.specs.DoseringsforslagRecordSpecs;
import com.trifork.stamdata.specs.SikredeRecordSpecs;
import com.trifork.stamdata.specs.VaccinationRecordSpecs;
import com.trifork.stamdata.specs.YderregisterRecordSpecs;

import dk.nsi.stamdata.jaxws.generated.Header;
import dk.nsi.stamdata.jaxws.generated.ObjectFactory;
import dk.nsi.stamdata.jaxws.generated.ReplicationFault;
import dk.nsi.stamdata.jaxws.generated.ReplicationRequestType;
import dk.nsi.stamdata.jaxws.generated.ReplicationResponseType;
import dk.nsi.stamdata.jaxws.generated.Security;
import dk.nsi.stamdata.jaxws.generated.StamdataReplication;
import dk.nsi.stamdata.jaxws.generated.StamdataReplicationService;
import dk.nsi.stamdata.replication.models.Client;
import dk.nsi.stamdata.replication.models.ClientDao;
import dk.nsi.stamdata.testing.TestHelper;
import dk.nsi.stamdata.security.DenGodeWebServiceFilter;
import dk.nsi.stamdata.testing.RequestHelper;
import dk.nsi.stamdata.testing.TestServer;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.EmptyCredentialVault;
import dk.sosi.seal.xml.XmlUtilException;

@RunWith(GuiceTestRunner.class)
public class StamdataReplicationImplIntegrationTest
{
    public static final String WHITELISTED_CVR = "25520041";
    public static final String NON_WHITELISTED_CVR = "87654321";
    private boolean isClientAuthority = false;

    private TestServer server;
    TestHelper helper = new TestHelper();
    private StamdataReplication client;

    private ReplicationRequestType request;
    private ReplicationResponseType response;
    private Element anyAsElement;

    @Inject
    private Session session;
    @Inject
    private ClientDao clientDao;
    @Inject
    private Provider<Connection> connectionProvider;

    private RecordSpecification recordSpecification = SikredeRecordSpecs.ENTRY_RECORD_SPEC;
    private List<Record> records = Lists.newArrayList();
    private DateTime now, lastYear, nextYear;
	private List<String> permissions;

	@Before
    public void setUp() throws Exception
    {
        server = new TestServer().port(8986).contextPath("/").start();

        URL wsdlLocation = new URL("http://localhost:8986/service/StamdataReplication?wsdl");
//        URL wsdlLocation = new URL("http://ext15-cniab01.nsp-test.netic.dk:8080/stamdata-batch-copy-ws/service/StamdataReplication?wsdl");
        QName serviceName = new QName("http://nsi.dk/2011/10/21/StamdataKrs/", "StamdataReplicationService");
        StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

        service.setHandlerResolver(new SealNamespaceResolver());
        client = service.getStamdataReplication();

        now = DateTime.now();
        lastYear = now.minusYears(1);
        nextYear = now.plusYears(1);

        isClientAuthority = true;

	    permissions = Arrays.asList(new String[] {
									"cpr/person/v1",
									"sikrede/sikrede/v1",
									"yderregister/yder/v3",
									"yderregister/person/v3",
									"bemyndigelsesservice/bemyndigelse/v1",
									"sor/sygehusafdeling/v1",
									"vitamin/grunddata/v1",
									"vitamin/firmadata/v1",
									"vitamin/udgaaedenavne/v1",
									"vitamin/indholdsstoffer/v1",
									"cpr/barnrelation/v1",
									"tilskudsblanket/forhoejettakst/v1",
									"ddv/vaccinationplan/v2",
									"doseringsforslag/dosageproposal/v2"});
    }

    @After
    public void tearDown() throws Exception
    {
    	if (server != null) {
    		server.stop();
    	}
    }


    @Test // SDM-85 Ualmindelig ValidFrom-dato returneres som dato med ugyldig tidszone
    public void testReallyOldValidFromDate() throws Exception {
        String validFrom = "0575-08-01 00:00:00";
        makeRecordWithSpecifiedValidFrom(validFrom);
        sendRequest();
        assertResponseContainsValueOnXPath("//yder:yder/yder:ValidFrom", "0575-08-02T23:00:00.000Z");
        assertAuditLog();
    }

    @Test
    public void testRecordElements() throws Exception {
        recordSpecification = YderregisterRecordSpecs.YDER_V3_RECORD_TYPE;
        Record record = new RecordBuilder(recordSpecification).field("HistIdYder", "1234567890123456").addDummyFieldsAndBuild();
        Record record2 = new RecordBuilder(recordSpecification).field("HistIdYder", "9876543210987654").addDummyFieldsAndBuild();
        records.add(record);
        records.add(record2);

        createReplicationRequest("yderregister", "yder", 3L);
        populateDatabaseAndSendRequest();
        assertAuditLog();
    }

    // Tests
    @Test
    public void testNotSoOldValidFromDate() throws Exception {
        // Efter 1894 skal vi ikke længere have utc
        String validFrom = "1894-08-01 00:00:00";
        makeRecordWithSpecifiedValidFrom(validFrom);
        sendRequest();
        assertResponseContainsValueOnXPath("//yder:yder/yder:ValidFrom", "1894-08-01T00:00:00.000+01:00");
        assertAuditLog();
    }

    private void makeRecordWithSpecifiedValidFrom(String validFrom) throws SQLException {
        recordSpecification = YderregisterRecordSpecs.YDER_V3_RECORD_TYPE;
        Record record = new RecordBuilder(YderregisterRecordSpecs.YDER_V3_RECORD_TYPE).
                field("HistIdYder", "1234567890123456").addDummyFieldsAndBuild();

        records.add(record);
        createReplicationRequest("yderregister", "yder", 3);

        Transaction t = session.beginTransaction();
        populateDatabaseNoCommit(true);
        Connection connection = session.connection();
        connection.createStatement().executeUpdate("UPDATE " + YderregisterRecordSpecs.YDER_V3_RECORD_TYPE.getTable() +
                " SET ValidFrom='"+ validFrom +"' WHERE HistIdYder='1234567890123456'");
        t.commit();
    }

    @Test(expected = SOAPFaultException.class)
    public void requestWithoutAnyReplicationTypeGivesSenderSoapFault() throws Exception
    {
        request = new ReplicationRequestType();
        populateDatabaseAndSendRequest();
    }

    @Test
    public void basicQueryDoesNotRaiseAnException() throws Exception
    {

        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        assertAuditLog();
    }

    @Ignore
    @Test(expected = ReplicationFault.class)
    public void cvrThatIsNotWhitelistedRaisesAnException() throws Exception
    {

        createSikredeSikredeRegisterReplicationRequest();
        isClientAuthority = false;
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void cvrThatIsWhitelistedForAnotherRegisterRaisesAnException() throws Exception
    {

        createSorHospitalRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void requestToAnUnexistingRegisterAndDatatypeRaisesAnException() throws Exception
    {

        createUnexistingRegisterAndDatatypeReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test(expected = ReplicationFault.class)
    public void requestToAnExistingRegisterButUnexistingDatatypeRaisesAnException() throws Exception
    {

        createExistingRegisterButUnexistingDatatypeReplicationRequest();
        populateDatabaseAndSendRequest();
    }

    @Test
    public void basicQueryOnEmptyDatabaseReturnsZeroResults() throws Exception
    {
        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        assertResponseContainsSikredeAtom();
        assertResponseContainsExactNumberOfRecords("sdm:sikrede", 0);
        assertAuditLog();
    }

    @Test
    public void testSikredeCopy() throws Exception
    {
        RecordBuilder builder = new RecordBuilder(SikredeRecordSpecs.ENTRY_RECORD_SPEC);
        builder.field("CPRnr", "1234567890");
        Record record = builder.addDummyFieldsAndBuild();
        records.add(record);

        createReplicationRequest("sikrede", "sikrede");
        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("sikrede", "sikrede");
        assertResponseContainsExactNumberOfRecords("sikrede:sikrede", 1);
        assertResponseContainsValueOnXPath("//sikrede:sikrede/sikrede:CPRnr", "1234567890");
        assertAuditLog();
    }

    @Test
    public void testYderregisterCopy() throws Exception
    {
        recordSpecification = YderregisterRecordSpecs.YDER_V3_RECORD_TYPE;
        Record record = new RecordBuilder(YderregisterRecordSpecs.YDER_V3_RECORD_TYPE).field("HistIdYder", "1234567890123456").addDummyFieldsAndBuild();
        records.add(record);

        createReplicationRequest("yderregister", "yder", 3L);

        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:HistIdYder", "1234567890123456");
        assertAuditLog();
    }

    @Test
    public void testYderregisterWhitelistningVersion() throws Exception
    {
         testYderregisterWhitelistning("yderregister/yder/v3");
    }

    @Test
    public void testYderregisterWhitelistningDatatype() throws Exception
    {
         testYderregisterWhitelistning("yderregister/yder/");
    }

    @Test
    public void testYderregisterWhitelistningRegister() throws Exception
    {
         testYderregisterWhitelistning("yderregister/");
    }

    @Test
    public void testYderregisterWhitelistningAll() throws Exception
    {
         testYderregisterWhitelistning("");
    }

    @Test
    public void testYderregisterWhitelistningVersionFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("yderregister/yder/v1");
        } catch (ReplicationFault fault) {
            assertEquals("The provided cvr (25520041) is not authorized to fetch this datatype (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    @Test
    public void testYderregisterWhitelistningDatatypeFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("yderregister/foo/");
        } catch (ReplicationFault fault) {
            assertEquals("The provided cvr (25520041) is not authorized to fetch this datatype (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    @Test
    public void testYderregisterWhitelistningRegisterFail() throws Exception
    {
        try {
            testYderregisterWhitelistning("foo");
        } catch (ReplicationFault fault) {
            assertEquals("The provided cvr (25520041) is not authorized to fetch this datatype (yderregister/yder/v3).",fault.getMessage());
            return;
        }
        fail();
    }

    public void testYderregisterWhitelistning(String permission) throws Exception
    {
        permissions = Arrays.asList(new String[] {permission});
        recordSpecification = YderregisterRecordSpecs.YDER_V3_RECORD_TYPE;
        Record record = new RecordBuilder(YderregisterRecordSpecs.YDER_V3_RECORD_TYPE).field("HistIdYder", "1234567890123456").addDummyFieldsAndBuild();
        records.add(record);

        createReplicationRequest("yderregister", "yder", 3L);

        populateDatabaseAndSendRequest();

        assertResponseContainsRecordAtom("yderregister", "yder", 3);
        assertResponseContainsExactNumberOfRecords("yder:yder", 1);
        assertResponseContainsValueOnXPath("//yder:yder/yder:HistIdYder", "1234567890123456");
        assertAuditLog();
    }


   @Test
   public void testYderregisterPersonCopy() throws Exception
   {
       recordSpecification = YderregisterRecordSpecs.PERSON_V3_RECORD_TYPE;
       Record record = new RecordBuilder(recordSpecification).field("HistIdPerson", "1234").addDummyFieldsAndBuild();
       records.add(record);

       createReplicationRequest("yderregister", "person", 3L);

       populateDatabaseAndSendRequest();

       assertResponseContainsRecordAtom("yderregister", "person", 3);
       assertResponseContainsExactNumberOfRecords("yder:person", 1);
       assertResponseContainsValueOnXPath("//yder:person/yder:HistIdPerson", "1234");
       assertAuditLog();
   }

    public String anyElementAsString() throws TransformerException
    {
        Transformer transformer = TransformerFactory.newInstance().newTransformer();
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        transformer.transform(new DOMSource(anyAsElement), new StreamResult(byteArrayOutputStream));
        return new String(byteArrayOutputStream.toByteArray());
    }

    @Test
    public void testThatWeAreAbleToHandleMoreThanTwentyRequestsForWsdlInSuccession() throws Exception
    {
        URL wsdlLocation = new URL("http://localhost:8986/service/StamdataReplication?wsdl");
        QName serviceName = new QName("http://nsi.dk/2011/10/21/StamdataKrs/", "StamdataReplicationService");

        for (int i = 0; i < 20; i++)
        {
            StamdataReplicationService service = new StamdataReplicationService(wsdlLocation, serviceName);

            service.setHandlerResolver(new SealNamespaceResolver());
            client = service.getStamdataReplication();
        }
    }

    @Test
    public void testThatWeAreAbleToHandleMoreThanTwentyConnectionsInSuccession() throws Exception
    {
        createSikredeSikredeRegisterReplicationRequest();
        populateDatabaseAndSendRequest();
        for (int i = 0; i < 20; i++)
        {
            sendRequest();
            assertAuditLog();
        }
    }

    @Test
    public void testCPRBarnRelationCopy() throws Exception
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("cpr");
        request.setDatatype("barnrelation");
        request.setVersion(1L);
        request.setOffset("00000000000000000000");
        request.setMaxRecords(500L);

        populateDatabaseAndSendRequest();

        //printDocument(anyAsElement.getOwnerDocument(), System.out);

        assertResponseContainsRecordAtom("cpr", "barnrelation");
        assertAuditLog();
    }

    @Test
    public void testDdvSDS3275() throws Exception
    {
        recordSpecification = VaccinationRecordSpecs.VACCINATIONPLANSV2_RECORD_SPEC;
        Record record = new RecordBuilder(recordSpecification).field("Id", "1234").addDummyFieldsAndBuild();
        records.add(record);

        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("ddv");
        request.setDatatype("vaccinationplan");
        request.setVersion(2L);
        request.setOffset("00000000000000000000");
        request.setMaxRecords(50L);

        populateDatabaseAndSendRequest();

        //printDocument(anyAsElement.getOwnerDocument(), System.out);

        assertResponseContainsRecordAtom("ddv", "vaccinationplan",2);
        assertAuditLog();
    }


	/**
	 * This is a regression test covering the change introduced by NSPSUPPORT-57 to the documentation
	 */
	@Test()
	public void testPermissionsAreNotCommaSeparated() throws Exception {
		createReplicationRequest("vitamin", "grunddata");

		// the documentation stated that this was the way to get the 4 vitamin permissions, so let's try it
		permissions = Arrays.asList("vitamin/grunddata/v1,vitamin/firmadata/v1,vitamin/udgaaedenavne/v1,vitamin/indholdsstoffer/v1");

		try {
			populateDatabaseAndSendRequest();
			fail("Expected SOAP fault saying that the client does not have the necessary permissions");
		} catch (ReplicationFault fault) {
			assertTrue("Fault message: " + fault.getMessage() , fault.getMessage().contains("The provided cvr (25520041) is not authorized to fetch this datatype (vitamin/grunddata/v1)"));
		}
	}

    @Test
    public void testViewNotFound() throws Exception {
        try {
            createReplicationRequest("sikrede", "sikrede-plappe");
            populateDatabaseAndSendRequest();
        } catch (ReplicationFault fault) {
            assertTrue(fault.getMessage().contains("Requested view: 'sikrede/sikrede-plappe/v1' not found"));
        }
    }

    @Test
    public void testXmlUtilException() throws Exception {

    		String fileName = "src/test/resources/data/invalidXml.xml";
    		String xml = RequestHelper.readFile(fileName);
    		String replication_endpoint = "http://localhost:8986/service/StamdataReplication";

    		String responseXml = RequestHelper.sendRequest(replication_endpoint, "http://nsi.dk/2011/10/21/StamdataKrs/replicate", xml);

    		String faultstring = RequestHelper.getFaultstring(responseXml);
    		assertEquals(faultstring.substring(0, 35), "Unable to parse XML at line number:");

    		String lineNumber = RequestHelper.getLineNumber(responseXml);
    		assertTrue(lineNumber.matches("\\d+"));

    		String columnNumber = RequestHelper.getColumnNumber(responseXml);
    		assertTrue(columnNumber.matches("\\d+"));

    		String message = RequestHelper.getMessage(responseXml);
    		assertEquals(message, "Element type \"foo\" must be followed by either attribute specifications, \"&gt;\" or \"/&gt;\".");
    }

    @Test
    public void testException() throws Exception {

    		String fileName = "src/test/resources/data/validXml.xml";
    		String xml = RequestHelper.readFile(fileName);
    		String replication_endpoint = "http://localhost:8986/service/StamdataReplication";

    		String responseXml = RequestHelper.sendRequest(replication_endpoint, "http://nsi.dk/2011/10/21/StamdataKrs/replicate", xml);

    		String faultstring = RequestHelper.getFaultstring(responseXml);
    		assertEquals(faultstring.substring(0, 24), "An exception was caught:");
    }

	// Helper methods

	// Pretty print XML document - good for debugging
    private static void printDocument(Document doc, OutputStream out) {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.transform(new DOMSource(doc), new StreamResult(new OutputStreamWriter(out, "UTF-8")));
        } catch(Exception e) {
            // ignore - this is a test method
            e.printStackTrace();
        }
    }

    private void createReplicationRequest(String register, String datatype, long version) {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister(register);
        request.setDatatype(datatype);
        request.setVersion(version);
        request.setOffset("0");
    }

    private void createReplicationRequest(String register, String datatype) {
    	createReplicationRequest(register, datatype, 1L);
    }

    private void createSikredeSikredeRegisterReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sikrede");
        request.setDatatype("sikrede");
        request.setVersion(1L);
        request.setOffset("0");
    }

    private void createSorHospitalRegisterReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sor");
        request.setDatatype("sygehus");
        request.setVersion(1L);
        request.setOffset("0");
    }

    private void createUnexistingRegisterAndDatatypeReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("foo");
        request.setDatatype("bar");
        request.setVersion(1L);
        request.setOffset("0");
    }

    private void createExistingRegisterButUnexistingDatatypeReplicationRequest()
    {
        request = new ObjectFactory().createReplicationRequestType();
        request.setRegister("sor");
        request.setDatatype("bar");
        request.setVersion(1L);
        request.setOffset("0");
    }

    private void assertResponseContainsSikredeAtom()
    {
        assertEquals(anyAsElement.getLocalName(), "feed");
        assertEquals(anyAsElement.getNamespaceURI(), "http://www.w3.org/2005/Atom");
        assertEquals(anyAsElement.getFirstChild().getFirstChild().getTextContent(), "tag:nsi.dk,2011:sikrede/sikrede/v1");
    }

    private void assertResponseContainsRecordAtom(String register, String datatype, int version)
    {
    	assertEquals(anyAsElement.getLocalName(), "feed");
    	assertEquals(anyAsElement.getNamespaceURI(), "http://www.w3.org/2005/Atom");
    	assertEquals(anyAsElement.getFirstChild().getFirstChild().getTextContent(), "tag:nsi.dk,2011:" + register + "/" + datatype + "/v"+version);
    }

    private void assertResponseContainsRecordAtom(String register, String datatype)
    {
    	assertResponseContainsRecordAtom(register, datatype, 1);
    }

    private void assertResponseContainsPersonWithSurNameMatchingGivenName(String givenName, String surName) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//sdm:person[sdm:fornavn='" + givenName + "']/sdm:efternavn");
        String result = expression.evaluate(anyAsElement);
        assertEquals(result, surName);
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, int n) throws XPathExpressionException
    {
        assertResponseContainsExactNumberOfRecords(tag, Integer.toString(n));
    }

    private void assertResponseContainsExactNumberOfRecords(String tag, String n) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("count(//" + tag + ")");
        String countAsString = expression.evaluate(anyAsElement);
        assertEquals(n, countAsString);
    }

    private void assertResponseContainsPersonWithCpr(String oracleCpr) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//sdm:person/sdm:cpr");
        String result = expression.evaluate(anyAsElement);

        //printDocument(anyAsElement.getOwnerDocument(), System.out);

        assertEquals(result, oracleCpr);
    }

    private void assertResponseContainsValueOnXPath(String xPath, String oracle) throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression(xPath);
        String result = expression.evaluate(anyAsElement);
        assertEquals(oracle, result);
    }

    private String getOffsetFromAtomFirstEntry() throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//atom:entry/atom:id");
        String atomId = expression.evaluate(anyAsElement);
        String[] tokens = atomId.split("/");
        return tokens[0];
    }

    private String getOffsetFromAtomLastEntry() throws XPathExpressionException
    {
        XPathExpression expression = createXpathExpression("//atom:entry/atom:id");
        String atomId = expression.evaluate(anyAsElement);
        String[] tokens = atomId.split("/");
        return tokens[tokens.length - 1];
    }

    private XPathExpression createXpathExpression(String expression) throws XPathExpressionException
    {
        NamespaceContext context = new NamespaceContextMap(
                "krs", "http://nsi.dk/2011/10/21/StamdataKrs/",
                "atom", "http://www.w3.org/2005/Atom",
                "sdm", "http://nsi.dk/-/stamdata/3.0/cpr",
                "sikrede", "http://nsi.dk/-/stamdata/3.0/sikrede",
                "yder", "http://nsi.dk/-/stamdata/3.0/yderregister",
                "grunddata", "http://nsi.dk/-/stamdata/3.0/vitamin",
                "forhoejettakst", "http://nsi.dk/-/stamdata/3.0/tilskudsblanket",
                "bemyndigelse", "http://nsi.dk/-/stamdata/3.0/bemyndigelsesservice",
        		"dosageproposal", "http://nsi.dk/-/stamdata/3.0/doseringsforslag",
        		"doseringsforslag", "http://nsi.dk/-/stamdata/3.0/doseringsforslag");
        XPathFactory factory = XPathFactory.newInstance();
        XPath xpath = factory.newXPath();
        xpath.setNamespaceContext(context);
        return xpath.compile(expression);
    }

    private void populateDatabaseAndSendRequest(boolean createTable) throws Exception
    {
        populateDatabase(createTable);
        sendRequest();
    }

    private void populateDatabaseAndSendRequest() throws Exception
    {
    	populateDatabaseAndSendRequest(true);
    }

    private void populateDatabase(boolean createTable) throws SQLException
    {
        Transaction t = session.beginTransaction();
        populateDatabaseNoCommit(createTable);
        t.commit();
    }

    private void populateDatabaseNoCommit(boolean createTable) throws SQLException {
        session.createQuery("DELETE FROM Client").executeUpdate();
        session.createSQLQuery("DELETE FROM Client_permissions").executeUpdate();

        // Example of subject serial number: CVR:19343634-UID:1234
        Client cvrClient = clientDao.create("Region Syd", String.format("CVR:%s-UID:1234", WHITELISTED_CVR));
        for (String permission : permissions) {
            cvrClient.addPermission(permission);
        }
        session.persist(cvrClient);

        Connection connection = session.connection();
        if (createTable) {
        	connection.createStatement().executeUpdate("DROP TABLE IF EXISTS " + recordSpecification.getTable());
        	connection.createStatement().executeUpdate(RecordMySQLTableGenerator.createSqlSchema(recordSpecification));
        } else {
        	connection.createStatement().executeUpdate("DELETE from " + recordSpecification.getTable());
        	connection.commit();
        }
        RecordPersister recordPersister = new RecordPersister(connectionProvider, new Instant());

        for (Record r : records) {
            recordPersister.persist(r, recordSpecification);
        }
    }

    private void sendRequest() throws Exception
    {
    	helper.enableAuditlogging(true);
        helper.redirectSystemOutToStream();

    	Holder<Security> securityHeader;
        Holder<Header> medcomHeader;

        if (isClientAuthority)
        {
            SecurityWrapper secutityHeadersNotWhitelisted = DGWSHeaderUtil.getVocesTrustedSecurityWrapper(WHITELISTED_CVR);
            securityHeader = secutityHeadersNotWhitelisted.getSecurity();
            medcomHeader = secutityHeadersNotWhitelisted.getMedcomHeader();
        } else {
            SecurityWrapper securityHeaders = DGWSHeaderUtil.getVocesTrustedSecurityWrapper(NON_WHITELISTED_CVR);
            securityHeader = securityHeaders.getSecurity();
            medcomHeader = securityHeaders.getMedcomHeader();
        }
        response = client.replicate(securityHeader, medcomHeader, request);
        anyAsElement = (Element) response.getAny();

        helper.resetSystemOut();
        helper.enableAuditlogging(false);

//        printDocument(anyAsElement.getOwnerDocument(), System.out);
    }

	public void assertAuditLog() throws XPathExpressionException {
		String[] logs = helper.getSystemLogs();
        NodeList elements = anyAsElement.getElementsByTagName("atom:entry");

        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"register\",\"value\":\"" + request.getRegister() + "\"}", logs[0]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"datatype\",\"value\":\"" + request.getDatatype() + "\"}", logs[1]);
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"version\",\"value\":\"" + request.getVersion() + "\"}", logs[2]);
        if (request.getMaxRecords() != null) {
        	assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"max records\",\"value\":\"" + request.getMaxRecords() + "\"}", logs[3]);
		} else {
			assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"max records\",\"value\":\"Undefined\"}", logs[3]);
		}
        assertEquals("DEVELOPMENT AUDIT: {\"component\":\"SDM\",\"context\":\"SKRS Replicate\",\"type\":\"SensitivePersonalInformation\",\"key\":\"offset\",\"value\":\"" + request.getOffset() + "\"}", logs[4]);
        assertEquals(elements.getLength(), Integer.parseInt(getValueFromLog(logs[5])));
        if (elements.getLength() > 0) {
        Node first = elements.item(0);
        Node last = elements.item(elements.getLength()-1);
        String firstRequest = first.getFirstChild().getTextContent().split("/")[3];
        String lastRequest = last.getFirstChild().getTextContent().split("/")[3];
        assertEquals(getValueFromLog(logs[6]), firstRequest);
        assertEquals(getValueFromLog(logs[7]), lastRequest);
        }
	}

	public String getValueFromLog(String log) {
		Scanner scanner = new Scanner(log);

        scanner.useDelimiter("\"");
        scanner.findInLine("\"value\":");
        String value = scanner.nextBigInteger().toString();
		return value;
	}

}
