/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.dao;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import com.google.inject.Inject;
import com.google.inject.Provider;

public class CprDAO {
    public enum WardRelation {
        BARN("barn"),
        BARN_UDEN("barn_uden"),
        MYNDLING("myndling");
        private final String value;

        WardRelation(String v) {
            value = v;
        }

        public String value() {
            return value;
        }

        public static WardRelation fromValue(String value) {
            for (WardRelation ward: WardRelation.values()) {
                if (ward.value.equals(value)) {
                    return ward;
                }
            }
            throw new IllegalArgumentException(value);
        }
    }

    public class Ward {
        String identifier;
        WardRelation relation;

        @Override
        public String toString() {
            return "Ward [identifier=" + identifier + ", relation=" + relation + "]";
        }
        public String getIdentifier() {
            return identifier;
        }
        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }
        public WardRelation getRelation() {
            return relation;
        }
        public void setRelation(WardRelation relation) {
            this.relation = relation;
        }
    }

    public enum CustodyRelation {
        MOR("mor"),
        MOR_UDEN("mor_uden"),
        FAR("far"),
        FAR_UDEN("far_uden"),
        ANDEN("anden"),
        VAERGE("vaerge");
        private final String value;

        CustodyRelation(String v) {
            value = v;
        }

        public String value() {
            return value;
        }

        public static CustodyRelation fromValue(String value) {
            for (CustodyRelation custody: CustodyRelation.values()) {
                if (custody.value.equals(value)) {
                    return custody;
                }
            }
            throw new IllegalArgumentException(value);
        }
    }

    public class Custody {
        String identifier;
        CustodyRelation relation;

        public String getIdentifier() {
            return identifier;
        }
        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }
        public CustodyRelation getRelation() {
            return relation;
        }
        public void setRelation(CustodyRelation relation) {
            this.relation = relation;
        }
        @Override
        public String toString() {
            return "Custody [identifier=" + identifier + ", relation=" + relation + "]";
        }
    }

    public enum SpouseRelation {
        AEGTEFAELLE("aegtefaelle"),
        REGISTRET_PARTNER("registreret_partner"),
        SAMBOENDE("samboende");
        private final String value;

        SpouseRelation(String v) {
            value = v;
        }

        public String value() {
            return value;
        }

        public static SpouseRelation fromValue(String value) {
            for (SpouseRelation spouse: SpouseRelation.values()) {
                if (spouse.value.equals(value)) {
                    return spouse;
                }
            }
            throw new IllegalArgumentException(value);
        }
    }

    public class Spouse {
        String identifier;
        String civilstand;
        SpouseRelation relation;

        public String getIdentifier() {
            return identifier;
        }

        public void setIdentifier(String identifier) {
            this.identifier = identifier;
        }

        public String getCivilstand() {
            return civilstand;
        }

        public void setCivilstand(String civilstand) {
            this.civilstand = civilstand;
            switch (civilstand) {
            case "G":
                relation = SpouseRelation.AEGTEFAELLE;
                break;
            case "U":
                relation = SpouseRelation.SAMBOENDE;
                break;
            case "P":
                relation = SpouseRelation.REGISTRET_PARTNER;
                break;
            default:
                relation = null;
                break;
            }
        }

        public SpouseRelation getRelation() {
            return relation;
        }
    }

    private final Provider<Connection> connectionProvider;

    @Inject
    public CprDAO(Provider<Connection> connectionProvider)
    {
        this.connectionProvider = connectionProvider;
    }

    public List<String> findRelations(String cpr) throws SQLException {
        return findRelations(cpr, new Date());
    }

    public List<String> findRelations(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "(select br.CPR from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where fmr.CPR = ? and fmr.TypeKode = '0003'"+
                                        " and p.Koen = 'K'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select br.CPR from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where fmr.CPR = ? and fmr.TypeKode = '0004'"+
                                        " and p.Koen = 'M'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select fmr.RelationCpr as CPR from v2_ForaeldreMyndighedRelation as fmr"+
                                 " where fmr.CPR = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006')"+
                                    " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null));";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            preparedStatement.setObject(6, cpr);
            preparedStatement.setObject(7, currentTime);
            preparedStatement.setObject(8, currentTime);
            preparedStatement.setObject(9, currentTime);
            preparedStatement.setObject(10, currentTime);
            preparedStatement.setObject(11, cpr);
            preparedStatement.setObject(12, currentTime);
            preparedStatement.setObject(13, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            List<String> relations = new ArrayList<String>();
            while (resultSet.next()) {
                relations.add(resultSet.getString("CPR"));
            }

            return relations;
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    public String findPriorCpr(String cpr) throws SQLException {
        return findPriorCpr(cpr, new Date());
    }

    public String findPriorCpr(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "select HistoriskCPR from HistoriskePersonnumre where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                return resultSet.getString("HistoriskCPR");
            } else {
                return null;
            }
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    public Date findMunicipalityCodeStartDate(String cpr) throws SQLException {
        return findMunicipalityCodeStartDate(cpr, new Date());
    }
    public Date findMunicipalityCodeStartDate(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "select TilflytningKommuneDato from PersonKontakt where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                return resultSet.getTimestamp("TilflytningKommuneDato");
            } else {
                return null;
            }
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    public Spouse findSpouse(String cpr) throws SQLException {
        return findSpouse(cpr, new Date());
    }

    public Spouse findSpouse(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "select Civilstand, AegtefaelleCPR from PersonFamilie where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            if (resultSet.next()) {
                Spouse spouse = new Spouse();
                spouse.setIdentifier(resultSet.getString("AegtefaelleCPR"));
                spouse.setCivilstand(resultSet.getString("Civilstand"));

                return spouse;

            } else {
                return null;
            }
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }

    }

    public Collection<Custody> findCustodies(String cpr) throws SQLException {
        return findCustodies(cpr, new Date());
    }

    public Collection<Custody> findCustodies(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "(select br.CPR, fmr.TypeKode, 'Mor' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where fmr.CPR = ? and fmr.TypeKode = '0003'"+
                                        " and p.Koen = 'K'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select br.CPR, fmr.TypeKode, 'Far' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where fmr.CPR = ? and fmr.TypeKode = '0004'"+
                                        " and p.Koen = 'M'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                    "(select br.CPR, '' as TypeKode, 'Mor_uden' as Relation from v2_BarnRelation as br"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.BarnCPR = ?"+
                                        " and p.Koen = 'K'"+
                                        " and (select CPR from v2_ForaeldreMyndighedRelation where CPR = ? and TypeKode = '0003'"+
                                          " and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)) is null"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select br.CPR, '' as TypeKode, 'Far_uden' as Relation from v2_BarnRelation as br"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.BarnCPR = ?"+
                                        " and p.Koen = 'M'"+
                                        " and (select CPR from v2_ForaeldreMyndighedRelation where CPR = ? and TypeKode = '0004'"+
                                          " and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)) is null"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select fmr.RelationCpr as CPR, fmr.TypeKode, 'Anden' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                   " join v2_Person as p on fmr.RelationCpr = p.CPR"+
                                   " where fmr.CPR = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006')"+
                                        " and p.Koen = 'K'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null))"+
                                 "union"+
                                 "(select fmr.RelationCpr as CPR, fmr.TypeKode, 'Anden' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                   " join v2_Person as p on fmr.RelationCpr = p.CPR"+
                                   " where fmr.CPR = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006')"+
                                        " and p.Koen = 'M'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null))"+
                                 "union"+
                                 "(select RelationCpr as CPR, '' as TypeKode, 'Vaerge' as Relation from v2_UmyndiggoerelseVaergeRelation"+
                                   " where CPR = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null));";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);
            preparedStatement.setObject(6, cpr);
            preparedStatement.setObject(7, currentTime);
            preparedStatement.setObject(8, currentTime);
            preparedStatement.setObject(9, currentTime);
            preparedStatement.setObject(10, currentTime);

            preparedStatement.setObject(11, cpr);
            preparedStatement.setObject(12, cpr);
            preparedStatement.setObject(13, currentTime);
            preparedStatement.setObject(14, currentTime);
            preparedStatement.setObject(15, currentTime);
            preparedStatement.setObject(16, currentTime);

            preparedStatement.setObject(17, cpr);
            preparedStatement.setObject(18, cpr);
            preparedStatement.setObject(19, currentTime);
            preparedStatement.setObject(20, currentTime);
            preparedStatement.setObject(21, currentTime);
            preparedStatement.setObject(22, currentTime);

            preparedStatement.setObject(23, cpr);
            preparedStatement.setObject(24, currentTime);
            preparedStatement.setObject(25, currentTime);

            preparedStatement.setObject(26, cpr);
            preparedStatement.setObject(27, currentTime);
            preparedStatement.setObject(28, currentTime);

            preparedStatement.setObject(29, cpr);
            preparedStatement.setObject(30, currentTime);
            preparedStatement.setObject(31, currentTime);

            ResultSet resultSet = preparedStatement.executeQuery();
            ArrayList<Custody> custodies = new ArrayList<Custody>();
            while (resultSet.next()) {
                Custody custody = new Custody();
                custody.setIdentifier(resultSet.getString("CPR"));
                String type = resultSet.getString("Relation");
                switch (type) {
                    case "Mor":
                        custody.setRelation(CustodyRelation.MOR);
                        break;
                    case "Far":
                        custody.setRelation(CustodyRelation.FAR);
                        break;
                    case "Mor_uden":
                        custody.setRelation(CustodyRelation.MOR_UDEN);
                        break;
                    case "Far_uden":
                        custody.setRelation(CustodyRelation.FAR_UDEN);
                        break;
                    case "Anden":
                        custody.setRelation(CustodyRelation.ANDEN);
                        break;
                    case "Vaerge":
                        custody.setRelation(CustodyRelation.VAERGE);
                        break;
                }
                custodies.add(custody);
            }

            return custodies;
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

    public Collection<Ward> findWard(String cpr) throws SQLException {
        return findWard(cpr, new Date());
    }

    public Collection<Ward> findWard(String cpr, Date currentTime) throws SQLException {
        PreparedStatement preparedStatement = null;
        Connection connection = null;
        try {
            String queryString = "(select br.BarnCPR, 'Barn' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.CPR = ? and fmr.TypeKode = '0003'"+
                                        " and p.Koen = 'K'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select br.BarnCPR, 'Barn' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                    " join v2_BarnRelation as br on fmr.CPR = br.BarnCPR"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.CPR = ? and fmr.TypeKode = '0004'"+
                                        " and p.Koen = 'M'"+
                                        " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null)"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                    "(select br.BarnCPR, 'Barn_uden' as Relation from v2_BarnRelation as br"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.CPR = ?"+
                                        " and p.Koen = 'K'"+
                                        " and (select CPR from v2_ForaeldreMyndighedRelation where CPR = br.BarnCPR and TypeKode = '0003'"+
                                          " and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)) is null"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select br.BarnCPR, 'Barn_uden' as Relation from v2_BarnRelation as br"+
                                    " join v2_Person as p on br.CPR = p.CPR"+
                                    " where br.CPR = ?"+
                                        " and p.Koen = 'M'"+
                                        " and (select CPR from v2_ForaeldreMyndighedRelation where CPR = br.BarnCPR and TypeKode = '0004'"+
                                          " and ValidFrom <= ? and (ValidTo > ? or ValidTo is null)) is null"+
                                        " and br.ValidFrom <= ? and (br.ValidTo > ? or br.ValidTo is null))"+
                                 "union"+
                                 "(select fmr.Cpr as BarnCPR, 'Barn' as Relation from v2_ForaeldreMyndighedRelation as fmr"+
                                 " where fmr.RelationCpr = ? and (fmr.TypeKode = '0005' OR fmr.TypeKode = '0006')"+
                                    " and fmr.ValidFrom <= ? and (fmr.ValidTo > ? or fmr.ValidTo is null))"+
                                 "union"+
                                 "(select CPR as BarnCPR, 'Myndling' as Relation from v2_UmyndiggoerelseVaergeRelation"+
                                    " where RelationCpr = ? and ValidFrom <= ? and (ValidTo > ? or ValidTo is null));";
            connection = connectionProvider.get();
            preparedStatement = connection.prepareStatement(queryString);
            preparedStatement.setObject(1, cpr);
            preparedStatement.setObject(2, currentTime);
            preparedStatement.setObject(3, currentTime);
            preparedStatement.setObject(4, currentTime);
            preparedStatement.setObject(5, currentTime);

            preparedStatement.setObject(6, cpr);
            preparedStatement.setObject(7, currentTime);
            preparedStatement.setObject(8, currentTime);
            preparedStatement.setObject(9, currentTime);
            preparedStatement.setObject(10, currentTime);
            preparedStatement.setObject(11, cpr);
            preparedStatement.setObject(12, currentTime);
            preparedStatement.setObject(13, currentTime);
            preparedStatement.setObject(14, currentTime);
            preparedStatement.setObject(15, currentTime);

            preparedStatement.setObject(16, cpr);
            preparedStatement.setObject(17, currentTime);
            preparedStatement.setObject(18, currentTime);
            preparedStatement.setObject(19, currentTime);
            preparedStatement.setObject(20, currentTime);

            preparedStatement.setObject(21, cpr);
            preparedStatement.setObject(22, currentTime);
            preparedStatement.setObject(23, currentTime);

            preparedStatement.setObject(24, cpr);
            preparedStatement.setObject(25, currentTime);
            preparedStatement.setObject(26, currentTime);
            ResultSet resultSet = preparedStatement.executeQuery();
            ArrayList<Ward> wards = new ArrayList<Ward>();
            while (resultSet.next()) {
                Ward ward = new Ward();
                ward.setIdentifier(resultSet.getString("BarnCPR"));
                String type = resultSet.getString("Relation");
                switch (type) {
                    case "Barn":
                        ward.setRelation(WardRelation.BARN);
                        break;
                    case "Barn_uden":
                        ward.setRelation(WardRelation.BARN_UDEN);
                        break;
                    case "Myndling":
                        ward.setRelation(WardRelation.MYNDLING);
                        break;
                }
                wards.add(ward);
            }

            return wards;
        } finally {
            if(preparedStatement != null) {
                preparedStatement.close();
            }
            if (connection != null) {
                connection.close();
            }
        }
    }

}
