/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.mapping.v103;

import java.math.BigInteger;
import java.util.Date;
import java.util.List;

import javax.xml.datatype.DatatypeConfigurationException;

import org.apache.commons.lang.StringUtils;

import com.google.inject.Inject;
import com.google.inject.servlet.RequestScoped;
import com.trifork.stamdata.Nullable;
import com.trifork.stamdata.Preconditions;
import com.trifork.stamdata.persistence.Record;

import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.cpr.mapping.CivilRegistrationStatusCodes;
import dk.nsi.stamdata.cpr.mapping.MunicipalityMapper;
import dk.nsi.stamdata.cpr.mapping.YderregisterRecordToAssociatedGeneralPractitionerMapper;
import dk.nsi.stamdata.cpr.mapping.PersonMapper.CPRProtectionLevel;
import dk.nsi.stamdata.cpr.medcom.FaultMessages;
import dk.nsi.stamdata.cpr.models.HealthCareInformation;
import dk.nsi.stamdata.cpr.models.Person;
import dk.nsi.stamdata.security.WhitelistService;
import dk.oio.rep.cpr_dk.xml.schemas.core._2005._11._24.PersonBirthDateStructureType;
import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.SimpleCPRPersonType;
import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.CountryIdentificationCodeType;
import dk.oio.rep.ebxml.xml.schemas.dkcc._2003._02._13.PersonGenderCodeType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.SystemIDCard;
import oio.medcom.cprservice._1_0_3.AddressAccessType;
import oio.medcom.cprservice._1_0_3.AddressCompleteType;
import oio.medcom.cprservice._1_0_3.AddressPostalType;
import oio.medcom.cprservice._1_0_3.FrontierCitizenshipStructureType;
import oio.medcom.cprservice._1_0_3.FrontierInformationStructureType;
import oio.medcom.cprservice._1_0_3.FrontierResidenceCountryStructureType;
import oio.medcom.cprservice._1_0_3.FrontierSocialSecurityAgenzyStructureType;
import oio.medcom.cprservice._1_0_3.FrontierSocialSecurityCountryStructureType;
import oio.medcom.cprservice._1_0_3.FrontierSocialSecurityStructureType;
import oio.medcom.cprservice._1_0_3.ObjectFactory;
import oio.medcom.cprservice._1_0_3.PersonAddressStructureType;
import oio.medcom.cprservice._1_0_3.PersonHealthCareInformationStructureType;
import oio.medcom.cprservice._1_0_3.PersonInformationStructureType;
import oio.medcom.cprservice._1_0_3.PersonPublicHealthInsuranceType;
import oio.medcom.cprservice._1_0_3.PersonWithHealthCareInformationStructureType;
import oio.medcom.cprservice._1_0_3.PriorGeneralPractitionerStructureType;
import oio.medcom.cprservice._1_0_3.PublicHealthInsuranceGroupIdentifierType;
import oio.medcom.cprservice._1_0_3.RegularCPRPersonType;
import oio.medcom.cprservice._1_0_3.PersonCivilRegistrationStatusStructureType;
import oio.medcom.cprservice._1_0_3.PersonCustodyStructureType;

/**
 * Person mapper for wsdl version 1.0.3
 */
@RequestScoped
public class PersonMapper extends dk.nsi.stamdata.cpr.mapping.PersonMapper {

	@Inject
    public PersonMapper(WhitelistService whitelistService, SystemIDCard idCard, MunicipalityMapper munucipalityMapper,
                        SLALogItem slaLogItem) {
        super(whitelistService, idCard, munucipalityMapper, slaLogItem);
    }

    public PersonInformationStructureType map(Person person, CPRProtectionLevel cprProtection, List<String> relations) {
        boolean censorData = isPersonProtected(person);
        return (censorData) ? createOutputWithCensoredDate(person, cprProtection) :
                createOutputWithRealData(person,relations);
    }
    
    // public PersonInformationStructureType

	private PersonInformationStructureType createOutputWithCensoredDate(
            Person person, CPRProtectionLevel cprProtection) {
        PersonInformationStructureType output = new ObjectFactory().createPersonInformationStructureType();
        mapCurrentPersonCivilRegistrationIdentifier(person, output);

        RegularCPRPersonType regularCprPerson = createCensoredRegularPerson103(person.getCpr(), cprProtection);

        // PERSON ADDRESS
        output.setPersonAddressStructure(createFakePersonAddressStructure(ADRESSEBESKYTTET));
        output.getPersonAddressStructure().setPersonInformationProtectionStartDate(newXMLGregorianCalendar(person.getNavnebeskyttelsestartdato()));
        output.setRegularCPRPerson(regularCprPerson);

        return output;
    }

    private PersonInformationStructureType createOutputWithRealData(Person person, List<String> relations) {
        // There are many cases in which we cannot fulfill the output format
        // just by using data present in the db. Therefore we have to 'fill'
        // in the gaps with e.g. 'UNKNOWN'.
        PersonInformationStructureType output = new ObjectFactory().createPersonInformationStructureType();

        mapCurrentPersonCivilRegistrationIdentifier(person, output);
        RegularCPRPersonType regularCprPerson = createRegularPerson103(person);

        // PERSON ADDRESS
        //
        // There are also records for which the record's values should not be
        // used. For instance if a person is dead or missing his address should
        // not be filled in. Yet the output schema requires them to be there.
        boolean shouldIncludeAddress = !CivilRegistrationStatusCodes.STATUSES_WITH_NO_ADDRESS.contains(person.getStatus());
        if (shouldIncludeAddress) {
            PersonAddressStructureType personAddress = new PersonAddressStructureType();

            if (person.navnebeskyttelsestartdato != null) {
                personAddress.setPersonInformationProtectionStartDate(newXMLGregorianCalendar(person.navnebeskyttelsestartdato));
            }

            // The output requires that the authority codes (municipality and
            // county) are exactly four char long.
            AddressCompleteType addressComplete = new AddressCompleteType();
            AddressAccessType addressAccess = new AddressAccessType();
            addressComplete.setAddressAccess(addressAccess);

            if (StringUtils.isNotBlank(person.kommuneKode)) {
                String municipalityCode = StringUtils.leftPad(person.kommuneKode, AUTHORITY_CODE_LENGTH, "0");
                addressAccess.setMunicipalityCode(municipalityCode);
                personAddress.setCountyCode(munucipalityMapper.toCountyCode(municipalityCode));
            } else {
                addressAccess.setMunicipalityCode("0000");
                addMutatedToSla("MunicipalityCode", "0000");
                personAddress.setCountyCode("0000");
                addMutatedToSla("CountyCode", "0000");
            }

            personAddress.setCareOfName(actualOrNull(person.coNavn));
            if (person.vejKode == null) {
                person.vejKode = "";
                addMutatedToSla("StreetCode", "0000");
            }
            addressAccess.setStreetCode(StringUtils.leftPad(person.vejKode, AUTHORITY_CODE_LENGTH, "0"));
            addressAccess.setStreetBuildingIdentifier(actualOrNull(person.husnummer));

            AddressPostalType addressPostal = new AddressPostalType();
            addressComplete.setAddressPostal(addressPostal);

            // The following field is not included in the source
            // "CPR Registeret"
            // therefore we cannot fill the element.
            //
            // if (StringUtils.isNotBlank(""))
            // {
            // This is:
            // The given name of a farm, estate, building or dwelling, which is
            // used as a additional postal address identifier.
            //
            // addressPostal.setMailDeliverySublocationIdentifier("Fake Value");
            // }

            addressPostal.setStreetName(actualOrUnknown("StreetName", person.vejnavn));
            addressPostal.setStreetNameForAddressingName(actualOrNull(person.getVejnavnTilAdressering()));
            addressPostal.setStreetBuildingIdentifier(actualOrNull(person.husnummer));

            addressPostal.setFloorIdentifier(actualOrNull(person.getEtage()));
            addressPostal.setSuiteIdentifier(actualOrNull(person.sideDoerNummer));

            // Documentation says:
            //
            // Name of a village, city or subdivision of a city or district,
            // which is determined as a part of the official address
            // specification for a certain street
            // or specific parts of a street, defined by intervals of street
            // building identifiers (da: house numbers).
            //
            // We believe that the CPR term for this is 'Lokalitet'.

            // KPN I think it should be bynavn ...
            addressPostal.setDistrictSubdivisionIdentifier(actualOrNull(person.lokalitet));

            // Post Box is excluded since a persons address cannot be a Post Box.
            if (person.postnummer == null) {
                addMutatedToSla("PostCodeIdentifier", "0000");
            }
            addressPostal.setPostCodeIdentifier(PersonMapper.normalizePostCode(person.postnummer));
            String postDistrict = PersonMapper.normalizePostDistrict(person.postdistrikt);
            if (!postDistrict.equals(person.postdistrikt)) {
                addMutatedToSla("DistrictName", postDistrict);
            }
            addressPostal.setDistrictName(postDistrict);

            CountryIdentificationCodeType countryIdentificationCode = mapCountryCodeFromStatus(person.status);
            if (countryIdentificationCode != null) addressPostal.setCountryIdentificationCode(countryIdentificationCode);

            personAddress.setAddressComplete(addressComplete);
            output.setPersonAddressStructure(personAddress);
        } else {
            output.setPersonAddressStructure(createFakePersonAddressStructure(UKENDT));
        }

        output.setRegularCPRPerson(regularCprPerson);

        for (String relation : relations) {
        	PersonCustodyStructureType personCustody = new PersonCustodyStructureType();
        	personCustody.setPersonCustodyIdentifier(relation);
        	output.getPersonCustodyStructure().add(personCustody);
        }

        return output;
    }


    public PersonWithHealthCareInformationStructureType map(Person person,
                                                            @Nullable HealthCareInformation healthCareInformation,
                                                            @Nullable Record yderRecord, @Nullable Record yderRecordGL, 
                                                            List<String> relations)
            throws DatatypeConfigurationException {
        Preconditions.checkNotNull(person, "person");

        PersonWithHealthCareInformationStructureType personWithHealthCare = new PersonWithHealthCareInformationStructureType();

        PersonInformationStructureType personInformation = map(person, CPRProtectionLevel.DoNotCensorCPR, relations);
        personWithHealthCare.setPersonInformationStructure(personInformation);

        PersonHealthCareInformationStructureType personHealthCareInformation = new PersonHealthCareInformationStructureType();
        personWithHealthCare.setPersonHealthCareInformationStructure(personHealthCareInformation);

        // Fill the associated general practitioner.
        AssociatedGeneralPractitionerStructureType associatedGeneralPractitioner = null;
        YderregisterRecordToAssociatedGeneralPractitionerMapper yderregisterRecordToAssociatedGeneralPractitionerMapper = new YderregisterRecordToAssociatedGeneralPractitionerMapper();
        if (yderRecord == null) {
            associatedGeneralPractitioner = createDummyPractitioner(UKENDT);
        } else {
            associatedGeneralPractitioner = yderregisterRecordToAssociatedGeneralPractitionerMapper.map(yderRecord, slaLogItem);
        }

        personHealthCareInformation.setAssociatedGeneralPractitionerStructure(associatedGeneralPractitioner);

        if (yderRecordGL != null) {
        	PriorGeneralPractitionerStructureType priorGeneralPractitioner = new PriorGeneralPractitionerStructureType();        
        	associatedGeneralPractitioner = yderregisterRecordToAssociatedGeneralPractitionerMapper.map(yderRecordGL, null);
        	priorGeneralPractitioner.setAssociatedGeneralPractitionerStructure(associatedGeneralPractitioner);
        	personHealthCareInformation.setPriorGeneralPractitionerStructure(priorGeneralPractitioner);
        }
        
        PersonPublicHealthInsuranceType personPublicHealthInsurance;
        if (healthCareInformation == null) {
            personPublicHealthInsurance = createDummyPublicHealthInsurance(UKENDT);
        } else {
            HealthCareGroupToPersonPublicHealthInsuranceMapper healthCareGroupToPersonPublicHealhInsuranceMapper =
                    new HealthCareGroupToPersonPublicHealthInsuranceMapper();
            personPublicHealthInsurance = healthCareGroupToPersonPublicHealhInsuranceMapper.map(healthCareInformation);
        }
        
        personHealthCareInformation.setPersonPublicHealthInsurance(personPublicHealthInsurance);
        
        if (validateHealthCareInformation(healthCareInformation)) {
            FrontierInformationStructureType frontierInformation = new FrontierInformationStructureType();

            frontierInformation.setFrontierStartDate(newXMLGregorianCalendar(healthCareInformation.getSSKGyldigFra()));
        	frontierInformation.setFrontierEndDate(newXMLGregorianCalendar(healthCareInformation.getSSKGyldigTil()));
        	        	
        	if (containsCitizenshipData(healthCareInformation)) {
        		FrontierCitizenshipStructureType citizenship = new FrontierCitizenshipStructureType();
        		
        		citizenship.setFrontierCitizenship(healthCareInformation.getSBSStatsborgerskab());
        		citizenship.setFrontierCitizenshipCode(healthCareInformation.getSBSStatsborgerskabKode());
        		
        		frontierInformation.setFrontierCitizenshipStructure(citizenship);
        	}
        	if (containsResidenceCountryData(healthCareInformation)) {
        		FrontierResidenceCountryStructureType residenceCountry = new FrontierResidenceCountryStructureType();
        		
        		residenceCountry.setFrontierResidenceCountry(healthCareInformation.getSSKBopelsLand());
        		residenceCountry.setFrontierResidenceCountryCode(healthCareInformation.getSSKBopelsLAndKode());
        		
        		frontierInformation.setFrontierResidenceCountryStructure(residenceCountry);
        	}
        	if (containsSocialSecurityData(healthCareInformation)) {
        		FrontierSocialSecurityStructureType socialSecurity = new FrontierSocialSecurityStructureType();
        		
        		FrontierSocialSecurityCountryStructureType ssCountry = new FrontierSocialSecurityCountryStructureType();
        		ssCountry.setFrontierSocialSecurityCountry(healthCareInformation.getSSLSocSikretLand());
        		ssCountry.setFrontierSocialSecurityCountryCode(healthCareInformation.getSSLSocSikretLandKode());
        		socialSecurity.setFrontierSocialSecurityCountryStructure(ssCountry);
        		
            	if (containsSocialSecurityAgenzyData(healthCareInformation)) {
            		FrontierSocialSecurityAgenzyStructureType ssAgenzy = new FrontierSocialSecurityAgenzyStructureType();
            		ssAgenzy.setFrontierSocialSecurityAgency(healthCareInformation.getSSLForsikringsinstans());
            		ssAgenzy.setFrontierSocialSecurityAgencyIdentifier(healthCareInformation.getSSLForsikringsinstansKode());
            		ssAgenzy.setFrontierSocialSecurityIdentifier(healthCareInformation.getSSLForsikringsnr());
            		socialSecurity.setFrontierSocialSecurityAgenzyStructure(ssAgenzy);
            	}
            	
        		frontierInformation.setFrontierSocialSecurityStructure(socialSecurity);
        	}

        	personWithHealthCare.setFrontierInformationStructure(frontierInformation);
        }
        
        
        return personWithHealthCare;
    }
    
	private boolean containsCitizenshipData(HealthCareInformation healthCareInformation) {
		return StringUtils.isNotEmpty(healthCareInformation.getSBSStatsborgerskab()) && 
			   StringUtils.isNotEmpty(healthCareInformation.getSBSStatsborgerskabKode());
	}

	private boolean containsResidenceCountryData(HealthCareInformation healthCareInformation) {
		return StringUtils.isNotEmpty(healthCareInformation.getSSKBopelsLand()) && 
			   StringUtils.isNotEmpty(healthCareInformation.getSSKBopelsLAndKode());
	}

	private boolean containsSocialSecurityData(HealthCareInformation healthCareInformation) {
		return StringUtils.isNotEmpty(healthCareInformation.getSSLSocSikretLand()) && 
			   StringUtils.isNotEmpty(healthCareInformation.getSSLSocSikretLandKode());
	}
	
	private boolean containsSocialSecurityAgenzyData(HealthCareInformation healthCareInformation) {
		return StringUtils.isNotEmpty(healthCareInformation.getSSLForsikringsinstans()) &&
			   StringUtils.isNotEmpty(healthCareInformation.getSSLForsikringsinstansKode()) &&
			   StringUtils.isNotEmpty(healthCareInformation.getSSLForsikringsnr());
	}

	private boolean validateHealthCareInformation(HealthCareInformation healthCareInformation) {
		return healthCareInformation != null && healthCareInformation.getSSKGyldigFra() != null && healthCareInformation.getSSKGyldigTil() != null;
	}

    public PersonPublicHealthInsuranceType createDummyPublicHealthInsurance(String placeholderText) throws DatatypeConfigurationException {
        Preconditions.checkNotNull(placeholderText, "placeholderText");
        PersonPublicHealthInsuranceType personPublicHealthInsurance = new PersonPublicHealthInsuranceType();
        personPublicHealthInsurance.setPublicHealthInsuranceGroupStartDate(newXMLGregorianCalendar(new Date(0)));
        personPublicHealthInsurance.setPublicHealthInsuranceGroupIdentifier(PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_1);
        return personPublicHealthInsurance;
    }

    private PersonAddressStructureType createFakePersonAddressStructure(String placeholderText)
    {
        PersonAddressStructureType personAddress = new PersonAddressStructureType();
        personAddress.setPersonInformationProtectionStartDate(null);

        personAddress.setCountyCode("0000");

        AddressCompleteType addressComplete = new AddressCompleteType();

        AddressAccessType addressAccess = new AddressAccessType();
        addressAccess.setMunicipalityCode("0000");
        addressAccess.setStreetCode("0000");

        addressComplete.setAddressAccess(addressAccess);
        AddressPostalType addressPostal = new AddressPostalType();
        addressComplete.setAddressPostal(addressPostal);

        addressPostal.setStreetName(placeholderText);
        addressPostal.setPostCodeIdentifier("0000");
        addressPostal.setDistrictName(placeholderText);

        personAddress.setAddressComplete(addressComplete);

        return personAddress;
    }

    private void mapCurrentPersonCivilRegistrationIdentifier(Person person, PersonInformationStructureType output) {
        if (StringUtils.isNotBlank(person.getGaeldendeCPR())) {
            output.setCurrentPersonCivilRegistrationIdentifier(person.getGaeldendeCPR());
        }
    }

    protected RegularCPRPersonType createCensoredRegularPerson103(String cpr, CPRProtectionLevel cprProtection) {
        RegularCPRPersonType regularCprPerson = new RegularCPRPersonType();
        SimpleCPRPersonType simpleCprPerson = new SimpleCPRPersonType();

        // PERSON NAME STRUCTURE SECTION

        PersonNameStructureType personName = new PersonNameStructureType();
        simpleCprPerson.setPersonNameStructure(personName);

        personName.setPersonGivenName(ADRESSEBESKYTTET);
        personName.setPersonMiddleName(null);
        personName.setPersonSurnameName(ADRESSEBESKYTTET);

        // For some searches we also need to censor the CPR number.

        boolean isCPRHidden = cprProtection == CPRProtectionLevel.CensorCPR;
        simpleCprPerson.setPersonCivilRegistrationIdentifier(isCPRHidden ? PROTECTED_CPR : cpr);

        regularCprPerson.setSimpleCPRPerson(simpleCprPerson);
        regularCprPerson.setPersonNameForAddressingName(ADRESSEBESKYTTET);

        // Even though the CPR can be read from the CPR (when the CPR is
        // included),
        // we might as well always protect it.
        regularCprPerson.setPersonGenderCode(PersonGenderCodeType.UNKNOWN);
        regularCprPerson.setPersonInformationProtectionIndicator(true);

        // BIRTH DATE
        PersonBirthDateStructureType personBirthDate = new PersonBirthDateStructureType();
        personBirthDate.setBirthDate(newXMLGregorianCalendar(new Date(0)));
        personBirthDate.setBirthDateUncertaintyIndicator(true);
        regularCprPerson.setPersonBirthDateStructure(personBirthDate);

        // CIVIL STATUS
        PersonCivilRegistrationStatusStructureType personCivil = new PersonCivilRegistrationStatusStructureType();
        personCivil.setPersonCivilRegistrationStatusCode(CivilRegistrationStatusCodes.AKTIV);
        personCivil.setPersonCivilRegistrationStatusStartDate(newXMLGregorianCalendar(new Date(0)));
        regularCprPerson.setPersonCivilRegistrationStatusStructure(personCivil);

        return regularCprPerson;
    }

    protected RegularCPRPersonType createRegularPerson103(Person person) {
        RegularCPRPersonType regularCprPerson = new RegularCPRPersonType();
        SimpleCPRPersonType simpleCprPerson = new SimpleCPRPersonType();
        regularCprPerson.setSimpleCPRPerson(simpleCprPerson);

        // PERSON NAME STRUCTURE SECTION
        //
        // Many of these values might not be present in the database
        // this is because that 'Such is life' and sometimes we simply
        // don't know a person's first name, address or birthday.
        PersonNameStructureType personName = new PersonNameStructureType();
        simpleCprPerson.setPersonNameStructure(personName);
        personName.setPersonGivenName(actualOrUnknown("PersonGivenName", person.fornavn));

        // Middle name is optional.
        personName.setPersonMiddleName(actualOrNull(person.mellemnavn));
        personName.setPersonSurnameName(actualOrUnknown("PersonSurnameName", person.efternavn));
        simpleCprPerson.setPersonCivilRegistrationIdentifier(person.cpr);

        regularCprPerson.setPersonNameForAddressingName(
                actualOrUnknown("PersonNameForAddressingName", person.getNavnTilAdressering()));
        regularCprPerson.setPersonGenderCode(mapGenderToGenderCode(person.koen));
        regularCprPerson.setPersonInformationProtectionIndicator(false);

        // BIRTH DATE
        PersonBirthDateStructureType personBirthDate = new PersonBirthDateStructureType();

        personBirthDate.setBirthDate(newXMLGregorianCalendar(person.foedselsdato));
        boolean foedselsdatoMarkering = false;
        if (person.getFoedselsdatoMarkering() != null) {
            String markering = person.getFoedselsdatoMarkering();
            if (markering.length() > 0 && !markering.equals("0")) {
                foedselsdatoMarkering = true;
            }
        }
        personBirthDate.setBirthDateUncertaintyIndicator(foedselsdatoMarkering);
        regularCprPerson.setPersonBirthDateStructure(personBirthDate);

        // CIVIL STATUS
        if (person.getStatus() == null) {
            // Vi kan ikke give noget fornuftigt svar hvis personen ikke engang har en status
            throw SoapUtils.newSOAPSenderFault(FaultMessages.INCOMPLETE_DATA_FOUND_FAULT_MSG);
        }
        PersonCivilRegistrationStatusStructureType personCivil = new PersonCivilRegistrationStatusStructureType();
        personCivil.setPersonCivilRegistrationStatusCode(person.getStatus());
        personCivil.setPersonCivilRegistrationStatusStartDate(newXMLGregorianCalendar(person.getStatusDato()));
        regularCprPerson.setPersonCivilRegistrationStatusStructure(personCivil);
        
        return regularCprPerson;
    }

}
