/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.medcom.v103;

import com.google.inject.Inject;
import com.sun.xml.ws.developer.SchemaValidation;
import com.trifork.stamdata.Fetcher;
import com.trifork.stamdata.jaxws.GuiceInstanceResolver.GuiceWebservice;
import com.trifork.stamdata.persistence.Record;
import com.trifork.stamdata.persistence.RecordFetcher;
import com.trifork.stamdata.persistence.Transactional;
import com.trifork.stamdata.specs.CprRecordSpecs.FMRTypeKode;
import dk.nsi.stamdata.cpr.SoapUtils;
import dk.nsi.stamdata.cpr.dao.CprDAO;
import dk.nsi.stamdata.cpr.dao.SikredeDAO;
import dk.nsi.stamdata.cpr.mapping.v103.PersonMapper;
import dk.nsi.stamdata.cpr.medcom.DetGodeCPROpslagCommon;
import dk.nsi.stamdata.cpr.medcom.ValidationHandler;
import dk.nsi.stamdata.cpr.models.HealthCareInformation;
import dk.nsi.stamdata.cpr.models.Person;
import dk.nsi.stamdata.cpr.util.UnreliableHandler;
import dk.nsi.stamdata.jaxws.generated.*;
import dk.nsi.stamdata.security.Whitelisted;
import dk.sdsd.nsp.slalog.api.SLALogItem;
import dk.sosi.seal.model.SystemIDCard;
import oio.medcom.cprservice._1_0_3.PersonInformationStructureType;
import oio.medcom.cprservice._1_0_3.DetGodeCPROpslag;
import oio.medcom.cprservice._1_0_3.GetPersonWithHealthCareInformationIn;
import oio.medcom.cprservice._1_0_3.GetPersonWithHealthCareInformationOut;
import oio.medcom.cprservice._1_0_3.DGWSFault;
import oio.medcom.cprservice._1_0_3.GetPersonInformationOut;
import oio.medcom.cprservice._1_0_3.GetPersonInformationIn;
import oio.medcom.cprservice._1_0_3.PersonWithHealthCareInformationStructureType;

import javax.jws.WebParam;
import javax.jws.WebService;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.ws.Holder;
import java.sql.SQLException;
import java.text.ParseException;
import java.util.Date;
import java.util.List;

@WebService(endpointInterface="oio.medcom.cprservice._1_0_3.DetGodeCPROpslag")
@GuiceWebservice
@SchemaValidation(handler = ValidationHandler.class)
public class DetGodeCPROpslagImpl extends DetGodeCPROpslagCommon implements DetGodeCPROpslag
{
    private static final String NS_DET_GODE_CPR_OPSLAG_103 = "urn:oio:medcom:cprservice:1.0.3";
    private static final String NS_DGWS_1_0 = "http://www.medcom.dk/dgws/2006/04/dgws-1.0.xsd";
    private static final String NS_WS_SECURITY = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";

    private final PersonMapper personMapper;
    private final SLALogItem slaLogItem;
	private CprDAO cpr2DAO;

    /**
     * Dummy
     */
    public DetGodeCPROpslagImpl() {
        super(null, null, null, null, null);
        slaLogItem = null;
        personMapper = null;
    }

    @Inject
    DetGodeCPROpslagImpl(Fetcher fetcher, RecordFetcher recordFetcher, PersonMapper personMapper, CprDAO foraeldreMyndighedDAO,
                         SikredeDAO sikredeDAO, SystemIDCard card, SLALogItem slaLogItem, UnreliableHandler unreliableHandler)
    {
        super(fetcher, recordFetcher, sikredeDAO, card.getSystemInfo().getCareProvider().getID(), unreliableHandler);
        this.personMapper = personMapper;
        this.slaLogItem = slaLogItem;
        this.cpr2DAO = foraeldreMyndighedDAO;
    }

    @Override
    @Whitelisted
    @Transactional
    public GetPersonWithHealthCareInformationOut getPersonWithHealthCareInformation(
            @WebParam(name = "Security", targetNamespace = NS_WS_SECURITY, mode = WebParam.Mode.INOUT, partName = "wsseHeader") Holder<Security> wsseHeader,
            @WebParam(name = "Header", targetNamespace = NS_DGWS_1_0, mode = WebParam.Mode.INOUT, partName = "medcomHeader") Holder<Header> medcomHeader,
            @WebParam(name = "getPersonWithHealthCareInformationIn", targetNamespace = NS_DET_GODE_CPR_OPSLAG_103, partName = "parameters") GetPersonWithHealthCareInformationIn input)
            throws DGWSFault
    {
        checkParseError(input, GetPersonWithHealthCareInformationIn.class);
        SoapUtils.updateSlaLog(medcomHeader, "getPersonWithHealthCareInformation", slaLogItem, clientCVR);
        SoapUtils.setHeadersToOutgoing(wsseHeader, medcomHeader);

        Date now = new Date();

        // 1. Check the white list to see if the client is authorized.
        String pnr = input.getPersonCivilRegistrationIdentifier();
        logAccess(pnr, "SCES v1.0.3 getPersonWithHealthcareInformation");

        // 2. Validate the input parameters.
        checkInputParameters(pnr);

        // 2. Fetch the person from the database.
        Person person = fetchPersonWithPnr(pnr);
        HealthCareInformation healthCareInformation = null;
        try {
            healthCareInformation = sikredeDAO.findHealthCareInformationWithFrontierInformation(pnr, now);
        } catch (SQLException e) {
            throw SoapUtils.newServerErrorFault(e);
        } catch (ParseException e) {
            throw SoapUtils.newServerErrorFault(e);
		}

        Record yderRecord = null;
        if (healthCareInformation != null) {
            try {
                yderRecord = getYderRecord(healthCareInformation.getProviderNo(), now);
            } catch (SQLException e) {
                throw SoapUtils.newServerErrorFault(e);
            }
        }
        Record yderRecordGL = null;
        if (healthCareInformation != null && healthCareInformation.getSYdernrGl() != null) {
            try {
				yderRecordGL = getYderRecord(healthCareInformation.getSYdernrGl(), now);
            } catch (SQLException e) {
                throw SoapUtils.newServerErrorFault(e);
            }
        }

        
        List<String> relations = null;
        try {
        	relations = cpr2DAO.findRelations(pnr,now);
        } catch (SQLException e) {
            throw SoapUtils.newServerErrorFault(e);
		}

        // If the relation is not found we have to use fake data to satisfy the
        // output format.
        GetPersonWithHealthCareInformationOut output = new GetPersonWithHealthCareInformationOut();
        PersonWithHealthCareInformationStructureType personWithHealthCareInformation = null;
        try {
            personWithHealthCareInformation = personMapper.map(person, healthCareInformation, yderRecord, yderRecordGL, relations);
        } catch (DatatypeConfigurationException e) {
            throw SoapUtils.newServerErrorFault(e);
        }

        output.setPersonWithHealthCareInformationStructure(personWithHealthCareInformation);

        return output;
    }

    @Override
    @Whitelisted
    @Transactional
    public GetPersonInformationOut getPersonInformation(
            @WebParam(name = "Security", targetNamespace = NS_WS_SECURITY, header = true, mode = WebParam.Mode.INOUT, partName = "wsseHeader") Holder<Security> wsseHeader,
            @WebParam(name = "Header", targetNamespace = NS_DGWS_1_0, header = true, mode = WebParam.Mode.INOUT, partName = "medcomHeader") Holder<Header> medcomHeader,
            @WebParam(name = "getPersonInformationIn", targetNamespace = NS_DET_GODE_CPR_OPSLAG_103, partName = "parameters") GetPersonInformationIn input) throws DGWSFault {

        checkParseError(input, GetPersonInformationIn.class);
        SoapUtils.updateSlaLog(medcomHeader, "getPersonInformation", slaLogItem, clientCVR);
        SoapUtils.setHeadersToOutgoing(wsseHeader, medcomHeader);

        // 1. Check the white list to see if the client is authorized.
        String pnr = input.getPersonCivilRegistrationIdentifier();
        logAccess(pnr, "SCES v1.0.3 getPersonInformation");

        // 2. Validate the input parameters.
        checkInputParameters(pnr);

        // 3. Fetch the person from the database.
        Person person = fetchPersonWithPnr(pnr);

        // We now have the requested person. Use it to fill in
        // the response.
        GetPersonInformationOut output = new GetPersonInformationOut();

        List<String> relations = null;
        try {
        	relations = cpr2DAO.findRelations(pnr,new Date());
        } catch (SQLException e) {
            throw SoapUtils.newServerErrorFault(e);
		}

        PersonInformationStructureType personInformation = personMapper.map(person, PersonMapper.CPRProtectionLevel.DoNotCensorCPR, relations);

        output.setPersonInformationStructure(personInformation);
        
        return output;
    }
}
