/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr;

import com.trifork.stamdata.specs.CprRecordSpecs;
import com.trifork.stamdata.specs.CprRecordSpecs.FMRTypeKode;
import com.trifork.stamdata.specs.SikredeRecordSpecs;
import com.trifork.stamdata.specs.YderregisterRecordSpecs;

import dk.nsi.stamdata.cpr.models.Person;

import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;

import com.trifork.stamdata.persistence.Record;
import com.trifork.stamdata.persistence.RecordBuilder;

import static com.trifork.stamdata.persistence.RecordSpecification.field;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Random;


public class Factories
{
    private static Random random = new Random();
    
    public static final Date YEAR_2100 = new DateTime(2100, 1, 1, 0, 0, 0).toDate();
    public static final Date YEAR_2000 = new DateTime(2000, 1, 1, 0, 0, 0).toDate();
    public static final Date YEAR_1999 = new DateTime(1999, 1, 1, 0, 0, 0).toDate();
    
    public static final Date TWO_DAYS_AGO = DateTime.now().minusDays(2).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).toDate();
    public static final Date YESTERDAY_PLUS_2_YEARS = DateTime.now().minusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).plusYears(2).toDate();
    public static final Date YESTERDAY = DateTime.now().minusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).toDate();
    public static final Date TOMORROW = DateTime.now().plusDays(1).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).toDate();
    public static final Date IN_TWO_DAYS = DateTime.now().plusDays(2).withHourOfDay(0).withMinuteOfHour(0).withSecondOfMinute(0).withMillisOfSecond(0).toDate();
    public static final Date FIRST_UNIX_DATE = new DateTime(1970, 1, 1, 0, 0, 0).toDate();

    public static String generateRandomCPR()
    {
        
        Calendar cal = Calendar.getInstance();
        cal.set(Calendar.YEAR, random.nextInt(98) + 1);
        cal.set(Calendar.MONTH, random.nextInt(11));
        cal.set(Calendar.DAY_OF_YEAR, random.nextInt(364) + 1);
        int no = random.nextInt(9999);

        return new SimpleDateFormat("ddMMyy").format(cal.getTime()) + String.format("%04d", no);
    }


    public static Person createPersonWithCPR(String cpr)
    {
        Person person = createPerson();
        person.setCpr(cpr);
        return person;
    }


    public static Person createPersonWithBirthday(Date birthday)
    {
        Person person = createPerson();
        person.setFoedselsdato(birthday);
        return person;
    }


    public static Person createPerson() {
        return createPerson(true, "6666");
    }

    public static Person createPerson(String postCode) {
        return createPerson(true, postCode);
    }

    public static Person createPerson(boolean hasHouseNo, String postCode)
    {
        Person person = new Person();

        person.setFornavn("Peter");
        person.setMellemnavn("Sigurd");
        person.setEfternavn("Andersen");
        person.setNavnTilAdressering("Peter,Andersen");

        person.setCpr(generateRandomCPR());
        
        // Yes, they are not identical.
        person.setGaeldendeCPR(generateRandomCPR());

        person.setKoen("M");

        person.setFoedselsdato(TWO_DAYS_AGO);

        person.setCoNavn("Søren Petersen");

        person.setKommuneKode("0461");
        person.setVejKode("0234");
        if (hasHouseNo) {
            person.setHusnummer("10");
        } else {
            person.setHusnummer("");
        }
        person.setBygningsnummer("A");
        person.setLokalitet("Birkely");
        person.setBynavn("BirkeBy");

        person.setVejnavn("Ørstedgade");
        person.setVejnavnTilAdressering("Østergd.");
        person.setEtage("12");
        person.setSideDoerNummer("tv");
        person.setFoedselsdatoMarkering("*");
        person.setStatus("01");
        person.setStatusDato(YESTERDAY);

        person.setPostnummer(postCode);
        person.setPostdistrikt("Überwald");

        person.setNavnebeskyttelsestartdato(null);
        person.setNavnebeskyttelseslettedato(null);

        person.setModifiedDate(TWO_DAYS_AGO);
        person.setCreatedDate(TWO_DAYS_AGO);
        person.setValidFrom(YESTERDAY);
        person.setValidTo(TOMORROW);

        return person;
    }


    public static Person createPersonWithAddressProtection()
    {
        Person person = createPerson();

        person.setNavnebeskyttelsestartdato(YESTERDAY);
        person.setNavnebeskyttelseslettedato(TOMORROW);

        return person;
    }

    public static Record createSikredeRecordFor(Person person, Record yderRecord, String groupCode, DateTime ikraftDato,
                                                DateTime iKraftYder)
    {
        RecordBuilder builder = new RecordBuilder(SikredeRecordSpecs.ENTRY_RECORD_SPEC);
        builder
                .field("CPRnr", person.getCpr())
                .field("SYdernr", yderRecord.get("YdernrYder"))
                .field("SSikrGrpKode", groupCode)
                .field("SIkraftDatoGrp", sikredeRecordDateString(ikraftDato))
                .field("SIkraftDatoYder", sikredeRecordDateString(iKraftYder))
        		.field("SSKGyldigFra", "")
        		.field("SSKGyldigTil", "")
        		;
        return builder.addDummyFieldsAndBuild();
    }

    public static Record createSikredeRecordFor(Person person, Record yderRecord, String groupCode, DateTime ikraftDato)
    {
        RecordBuilder builder = new RecordBuilder(SikredeRecordSpecs.ENTRY_RECORD_SPEC);
        builder
            .field("CPRnr", person.getCpr())
            .field("SYdernr", yderRecord.get("YdernrYder"))
            .field("SStatus", "D")
            .field("SSikrGrpKode", groupCode)
            .field("SIkraftDatoGrp", sikredeRecordDateString(ikraftDato))
            .field("SSKGyldigFra", "")
            .field("SSKGyldigTil", "");
        return builder.addDummyFieldsAndBuild();
    }

    public static Record createSikredeRecordFor(Person person, Record yderRecord, Record yderRecordGL, String groupCode, DateTime ikraftDato)
    {
        RecordBuilder builder = new RecordBuilder(SikredeRecordSpecs.ENTRY_RECORD_SPEC);
        builder
            .field("CPRnr", person.getCpr())
            .field("SYdernr", yderRecord.get("YdernrYder"))
            .field("SYdernrGl", yderRecordGL.get("YdernrYder"))
            .field("SStatus", "D")
            .field("SSikrGrpKode", groupCode)
            .field("SIkraftDatoGrp", sikredeRecordDateString(ikraftDato))
            .field("SSKGyldigFra", "")
            .field("SSKGyldigTil", "");
        return builder.addDummyFieldsAndBuild();
    }

    public static Record createForaeldreMyndighedRelationRecordFor(Person person, FMRTypeKode typeKode, Person relation)
    {
        RecordBuilder builder = new RecordBuilder(CprRecordSpecs.FORAELDRE_MYNDIGHED_RELATION_RECORD_TYPE);
        String relationCPR = (relation != null) ? relation.getCpr() : null;
        builder
        	.field("Id", person.getCpr()+"-"+typeKode.getId(relationCPR))
            .field("CPR", person.getCpr())
            .field("TypeKode", typeKode.getKode())
            .field("TypeTekst", typeKode.getTekst())
            .field("RelationCpr", relationCPR)
            .field("CreatedDate", "2000-01-01")
        	.field("ValidTo", "2050-01-01");

        return builder.addDummyFieldsAndBuild();
    }

    public static Record createBarnRelationRecordFor(Person person, Person barn)
    {
        RecordBuilder builder = new RecordBuilder(CprRecordSpecs.BARN_RELATION_RECORD_TYPE);
        builder
        	.field("Id", person.getCpr()+"-"+barn.getCpr())
            .field("CPR", person.getCpr())
            .field("BarnCPR", barn.getCpr())            
            .field("CreatedDate", "2000-01-01")
            .field("ValidTo", "2050-01-01");

        return builder.addDummyFieldsAndBuild();
    }

    public static Record createSikredeRecordWithFrontierInfo(Person person, Record yderRecord, String groupCode, DateTime ikraftDato,
            DateTime iKraftYder, boolean includeFrontierInfo, boolean includeSocialSecurity, boolean includeSocialSecurityAgenzy,boolean includeOpholdsland)
    {
    	RecordBuilder builder = new RecordBuilder(SikredeRecordSpecs.ENTRY_RECORD_SPEC);
    	builder
    	.field("CPRnr", person.getCpr())
    	.field("SYdernr", yderRecord.get("YdernrYder"))
    	.field("SSikrGrpKode", groupCode)
    	.field("SIkraftDatoGrp", sikredeRecordDateString(ikraftDato))
    	.field("SIkraftDatoYder", sikredeRecordDateString(iKraftYder))
    	.field("SSKGyldigFra", sikredeRecordFrontierDateString(ikraftDato))
    	.field("SSKGyldigTil", sikredeRecordFrontierDateString(ikraftDato.plusYears(2)))
    	;
    	if (includeFrontierInfo) {
    		String sskBopelsLand = "";
    		String sskBopelsLAndKode = "";
    		String sslSocSikretLand = "";
			String sslSocSikretLandKode = "";
			String sslForsikringsinstans = "";
			String sslForsikringsinstansKode = "";
			String sslForsikringsnr = "";
			if (includeSocialSecurity) {
				sslSocSikretLand = "Danmark";
				sslSocSikretLandKode = "DK";
				if (includeSocialSecurityAgenzy) {
					sslForsikringsinstans = "FOO sikring";
					sslForsikringsinstansKode = "BAR";
					sslForsikringsnr = "12345";
				}
			}
			if (includeOpholdsland) {
	    		sskBopelsLand = "Tyskland";
	    		sskBopelsLAndKode = "DE";				
			}
			builder
    		.field("SBSStatsborgerskab", "England")
    		.field("SBSStatsborgerskabKode", "GB")
    		.field("SSKBopelsLand", sskBopelsLand)
    		.field("SSKBopelsLAndKode", sskBopelsLAndKode)
    		.field("SSLSocSikretLand", sslSocSikretLand)
    		.field("SSLSocSikretLandKode", sslSocSikretLandKode)
    		.field("SSLForsikringsinstans", sslForsikringsinstans)
    		.field("SSLForsikringsinstansKode", sslForsikringsinstansKode)
   			.field("SSLForsikringsnr", sslForsikringsnr);
    	} else {
    		builder
    		.field("SBSStatsborgerskab", "")
    		.field("SBSStatsborgerskabKode", "")
    		.field("SSKBopelsLand", "")
    		.field("SSKBopelsLAndKode", "")
    		.field("SSLSocSikretLand", "")
    		.field("SSLSocSikretLandKode", "")
    		.field("SSLForsikringsinstans", "")
    		.field("SSLForsikringsinstansKode", "")
    		.field("SSLForsikringsnr", "")
    		;    		
    	}
    	return builder.addDummyFieldsAndBuild();
    }

    public static Record createYderRecord(String ydernummer)
    {
        RecordBuilder builder = new RecordBuilder(YderregisterRecordSpecs.YDER_V3_RECORD_TYPE)
                                    .field("Id", "langmd5her...")
                                    .field("YdernrYder", ydernummer)
                                    .field("PrakBetegn", "Klinikken")
                                    .field("PostdistYder", "Århus")
                                    .field("EmailYder", "test@example.com")
                                    .field("PostnrYder", "8000")
                                    .field("AdrYder", "Margrethepladsen 44, 8000 Århus")
                                    .field("HvdTlf", "12345678")
                                    .field("CVR",  "12345678");
        return builder.addDummyFieldsAndBuild();
    }
    
    private static String sikredeRecordDateString(DateTime date) 
    {
        DateTimeFormatter formatter = DateTimeFormat.forPattern("yyyyMMdd");
        return formatter.print(date);
    }

    private static String sikredeRecordFrontierDateString(DateTime date) 
    {
        DateTimeFormatter formatter = DateTimeFormat.forPattern("yyyy-MM-dd");
        return formatter.print(date);
    }
}
