/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr;

import static dk.nsi.stamdata.cpr.Factories.YESTERDAY;
import static dk.nsi.stamdata.cpr.mapping.v100.PersonMapper.newXMLGregorianCalendar;
import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertThat;

import java.math.BigInteger;
import java.util.Collections;
import java.util.Date;

import javax.xml.datatype.DatatypeConfigurationException;

import dk.nsi.stamdata.cpr.mapping.v100.PersonMapper;
import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.AssociatedGeneralPractitionerStructureType;
import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.PersonPublicHealthInsuranceType;
import dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.PersonWithHealthCareInformationStructureType;

import static dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_1;
import static dk.oio.rep.medcom_sundcom_dk.xml.schemas._2007._02._01.PublicHealthInsuranceGroupIdentifierType.SYGESIKRINGSGRUPPE_2;

import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;

import com.trifork.stamdata.persistence.Record;

import dk.nsi.stamdata.cpr.mapping.MunicipalityMapper;
import dk.nsi.stamdata.cpr.models.HealthCareInformation;
import dk.nsi.stamdata.cpr.models.Person;
import dk.nsi.stamdata.testing.MockSecureTokenService;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.SystemIDCard;


public class PersonHealthCareInfoMapperTest {
    private static final String NOT_WHITELISTED_CVR = "99999999";
    private static final String ADRESSEBESKYTTET = "ADRESSEBESKYTTET";
    private static final String UKENDT = "UKENDT";

    private Person person;
    private Record yderRecord;
    private HealthCareInformation healthCareGroup;

    private PersonWithHealthCareInformationStructureType output;


    @Before
    public void setUp() {
        person = Factories.createPerson();
        yderRecord = Factories.createYderRecord("1234");
        healthCareGroup = new HealthCareInformation(new DateTime(Factories.YESTERDAY).toDate(), "2", "1234");
    }

    private void doMapping() throws DatatypeConfigurationException {
        MunicipalityMapper municipalityMapper = new MunicipalityMapper();
        SystemIDCard idCard = MockSecureTokenService.createSignedSystemIDCard(NOT_WHITELISTED_CVR, AuthenticationLevel.VOCES_TRUSTED_SYSTEM);

        output = new PersonMapper(new StubWhitelistService(Collections.<String>emptyList()), idCard, municipalityMapper, null).
                map(person, healthCareGroup, yderRecord);
    }


    @Test
    public void itInsertsRealDataIfNoneAreMissing() throws DatatypeConfigurationException {
        doMapping();

        assertThatRelationIsRealData();
        assertThatGeneralPractitionerIsRealData();
    }


    @Test
    public void itInsertsDummyDataIfTheYderDataIsMissing() throws DatatypeConfigurationException {
        yderRecord = null;

        doMapping();

        assertThatRelationIsRealData();
        assertThatGeneralPractitionerIsDummy(UKENDT);
    }


    @Test
    public void itInsertsDummyDataIfThePersonIsProtectedAndDataIsMissing() throws DatatypeConfigurationException {
        person = Factories.createPersonWithAddressProtection();
        yderRecord = null;
        healthCareGroup = null;

        doMapping();

        assertThatRelationIsDummy();
    }


    @Test
    public void itInsertsDummyDataIfBothAreMissing() throws DatatypeConfigurationException {
        healthCareGroup = null;
        yderRecord = null;

        doMapping();

        assertThatRelationIsDummy();
        assertThatGeneralPractitionerIsDummy(UKENDT);
    }


    private void assertThatGeneralPractitionerIsDummy(String placeholderText) {
        AssociatedGeneralPractitionerStructureType g = output.getPersonHealthCareInformationStructure().getAssociatedGeneralPractitionerStructure();

        assertThat(g.getAssociatedGeneralPractitionerOrganisationName(), is(placeholderText));
        assertThat(g.getDistrictName(), is(placeholderText));
        assertThat(g.getStandardAddressIdentifier(), is(placeholderText));
        assertThat(g.getAssociatedGeneralPractitionerIdentifier(), is(BigInteger.ZERO));
        assertThat(g.getTelephoneSubscriberIdentifier(), is("00000000"));
        assertThat(g.getPostCodeIdentifier(), is("0000"));
        assertThat(g.getEmailAddressIdentifier(), is(placeholderText + "@example.com"));
    }


    private void assertThatGeneralPractitionerIsRealData() {
        AssociatedGeneralPractitionerStructureType g = output.getPersonHealthCareInformationStructure().getAssociatedGeneralPractitionerStructure();

        assertThat(g.getAssociatedGeneralPractitionerOrganisationName(), is(yderRecord.get("PrakBetegn")));
        assertThat(g.getDistrictName(), is(yderRecord.get("PostdistYder")));
        assertThat(g.getStandardAddressIdentifier(), is(yderRecord.get("AdrYder")));
        assertThat(g.getAssociatedGeneralPractitionerIdentifier(), is(new BigInteger((String) yderRecord.get("YdernrYder"))));
        assertThat(g.getTelephoneSubscriberIdentifier(), is(yderRecord.get("HvdTlf")));
        assertThat(g.getPostCodeIdentifier(), is(yderRecord.get("PostnrYder")));
        assertThat(g.getEmailAddressIdentifier(), is(yderRecord.get("EmailYder")));
    }


    private void assertThatRelationIsDummy() throws DatatypeConfigurationException {
        PersonPublicHealthInsuranceType g = output.getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance();

        assertThat(g.getPublicHealthInsuranceGroupIdentifier(), is(SYGESIKRINGSGRUPPE_1));
        assertThat(g.getPublicHealthInsuranceGroupStartDate(), is(newXMLGregorianCalendar(new Date(0))));
    }


    private void assertThatRelationIsRealData() throws DatatypeConfigurationException {
        PersonPublicHealthInsuranceType g = output.getPersonHealthCareInformationStructure().getPersonPublicHealthInsurance();

        assertThat(g.getPublicHealthInsuranceGroupIdentifier(), is(SYGESIKRINGSGRUPPE_2));
        assertThat(g.getPublicHealthInsuranceGroupStartDate().toGregorianCalendar().getTime(), is(YESTERDAY));
    }
}
