/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2011 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.stamdata.cpr.medcom;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.inject.Provider;
import com.trifork.stamdata.persistence.Record;
import com.trifork.stamdata.persistence.RecordBuilder;
import com.trifork.stamdata.persistence.RecordPersister;
import com.trifork.stamdata.specs.CprRecordSpecs;
import com.trifork.stamdata.specs.SikredeRecordSpecs;
import com.trifork.stamdata.specs.YderregisterRecordSpecs;
import com.trifork.stamdata.specs.CprRecordSpecs.FMRTypeKode;

import dk.nsi.stamdata.cpr.Factories;
import dk.nsi.stamdata.cpr.models.Person;

import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.joda.time.DateTime;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Map;

public class SCESTestHelper {
    private List<Person> persons = Lists.newArrayList();
    private List<Record> yderRecords = Lists.newArrayList();
    private List<Record> sikredeRecords = Lists.newArrayList();
    private List<Record> fmrRecords = Lists.newArrayList();
    private List<Record> brRecords = Lists.newArrayList();
    private Map<String,String> familieMap = Maps.newHashMap();
    private Map<String,String> seperatedMap = Maps.newHashMap();
    private List<String> unmarriedList = Lists.newArrayList();
    private Map<String,String> vaergeMap = Maps.newHashMap();
    public static final String CPR_VAERGE = "0102800005";
    public static final String CPR_MEDMOR = "0102800004";
    public static final String CPR_MEDFAR = "0102800003";
    public static final String CPR_MOR = "0102800002";
    public static final String CPR_FAR = "0102800001";
    public static final String CPR_BARN4 = "1111110004";
    public static final String CPR_BARN3 = "1111110003";
    public static final String CPR_BARN2 = "1111110002";
    public static final String CPR_BARN1 = "1111110001";
    public static final String CPR_SPOUSE = "0101014321";
    public static final String PRIOR_CPR = "0102030405";
    public static final String CPR_MISSING = "0102800006";

    SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd");

    public Date formatSimpleDate(String date) throws ParseException {
        return dateFormatter.parse(date);
    }

    public void addPerson(Person person) {
        persons.add(person);
    }

    public void addYder(Record yder) {
        yderRecords.add(yder);
    }

    public void addSikrede(Record sikrede) {
        sikredeRecords.add(sikrede);
    }

    public void addFarMorRelation(Record farMorReleation) {
        fmrRecords.add(farMorReleation);
    }

    public void addBarnRelation(Record barnRelation) {
        brRecords.add(barnRelation);
    }

    public Person createPerson() {
        Person person = Factories.createPerson();
        addPerson(person);

        return person;
    }

    public Person createPersonWithPostcode(String postcode) {
        Person person = Factories.createPerson(postcode);
        addPerson(person);

        return person;
    }

    public Person createPerson(String cpr) {
        Person person = Factories.createPersonWithCPR(cpr);
        addPerson(person);

        return person;
    }

    public Person createMand(String cpr) {
        Person person = createPerson(cpr);
        person.setKoen("M");

        return person;
    }

    public Person createKvinde(String cpr) {
        Person person = createPerson(cpr);
        person.setKoen("K");

        return person;
    }

    public Record createYder(String nummer) {
        Record yderRecord = Factories.createYderRecord(nummer);
        yderRecords.add(yderRecord);

        return yderRecord;
    }

    public Record createSikrede(Person person, Record yder, String groupCode, DateTime ikraftDato, DateTime iKraftYder) {
        Record sikredeRecord = Factories.createSikredeRecordFor(person, yder, groupCode, ikraftDato, iKraftYder);
        sikredeRecords.add(sikredeRecord);

        return sikredeRecord;
    }

    public Record createSikrede(Person person, Record yder, String groupCode, DateTime ikraftDato)
    {
        Record sikredeRecord = Factories.createSikredeRecordFor(person, yder, groupCode, ikraftDato);
        sikredeRecords.add(sikredeRecord);

        return sikredeRecord;
    }

    public Record createSikrede(Person person, Record yder, Record yderGL, String groupCode, DateTime ikraftDato) {
        Record sikredeRecord = Factories.createSikredeRecordFor(person, yder, yderGL, groupCode, ikraftDato);
        sikredeRecords.add(sikredeRecord);

        return sikredeRecord;
    }

    public Record createSikredeWithFrontierInfo(Person person, Record yder,String groupCode, DateTime ikraftDato, DateTime iKraftYder) {
        return createSikredeWithFrontierInfo(person, yder, groupCode, ikraftDato, iKraftYder, true, true, true, true);
    }

    public Record createSikredeWithFrontierInfo(Person person, Record yder,String groupCode, DateTime ikraftDato,
            DateTime iKraftYder, boolean includeFrontierInfo) {
        return createSikredeWithFrontierInfo(person, yder, groupCode, ikraftDato, iKraftYder, includeFrontierInfo, true, true, true);
    }

    public Record createSikredeWithFrontierInfo(Person person, Record yder,String groupCode, DateTime ikraftDato,
            DateTime iKraftYder, boolean includeFrontierInfo, boolean includeSocialSecurity) {
        return createSikredeWithFrontierInfo(person, yder, groupCode, ikraftDato, iKraftYder, includeFrontierInfo, includeSocialSecurity, true, true);
    }

    public Record createSikredeWithFrontierInfo(Person person, Record yder,String groupCode, DateTime ikraftDato,
            DateTime iKraftYder, boolean includeFrontierInfo, boolean includeSocialSecurity, boolean includeSocialSecurityAgenzy) {
        return createSikredeWithFrontierInfo(person, yder, groupCode, ikraftDato, iKraftYder, includeFrontierInfo, includeSocialSecurity, includeSocialSecurityAgenzy, true);
    }

    public Record createSikredeWithFrontierInfo(Person person, Record yder,String groupCode, DateTime ikraftDato,
            DateTime iKraftYder, boolean includeFrontierInfo, boolean includeSocialSecurity, boolean includeSocialSecurityAgenzy,boolean includeOpholdsland) {
        Record sikredeRecord = Factories.createSikredeRecordWithFrontierInfo(person, yder, groupCode, ikraftDato,iKraftYder,includeFrontierInfo, includeSocialSecurity, includeSocialSecurityAgenzy, includeOpholdsland);
        sikredeRecords.add(sikredeRecord);

        return sikredeRecord;
    }

    public Person createPersonWithAddressProtection() {
        Person person = Factories.createPersonWithAddressProtection();
        addPerson(person);

        return person;
    }

    public Record createForaeldreMyndighedRelation(Person person, FMRTypeKode type) {
        Record record = Factories.createForaeldreMyndighedRelationRecordFor(person, type,null);
        fmrRecords.add(record);

        return record;
    }

    public Record createForaeldreMyndighedRelation(Person person, FMRTypeKode type, Person relation) {
        Record record = Factories.createForaeldreMyndighedRelationRecordFor(person, type, relation);
        fmrRecords.add(record);

        return record;
    }

    public void createVaerge(Person person, Person vaerge) {
        vaergeMap.put(person.getCpr(), (vaerge == null) ? "" : vaerge.getCpr());
    }

    public Record createBarnRelation(Person person, Person barn) {
        Record record = Factories.createBarnRelationRecordFor(person, barn);
        brRecords.add(record);

        return record;
    }

    public void createSpouse(Person person, Person spouse) {
        familieMap.put(person.getCpr(), spouse.getCpr());
    }

    public void createSeperated(Person person, Person spouse) {
        seperatedMap.put(person.getCpr(), spouse.getCpr());
    }

    public void createUnmarried(Person person) {
        unmarriedList.add(person.getCpr());
    }

    public void deleteAllPersons(Session session) {
        Transaction tx = null;
        try {
            tx = session.beginTransaction();
            session.createQuery("DELETE FROM Person").executeUpdate();
            for (Person person : persons) {
                session.persist(person);
            }
            tx.commit();
        } catch (HibernateException e) {
            if (tx != null) {
                tx.rollback();
            }
            throw e;
        }
    }

    public Person findPerson(String cpr) {
        for (Person person : persons) {
            if (person.getCpr().equals(cpr)) {
                return person;
            }
        }
        return null;
    }

    public void prepareTestdata(Session session, Provider<Connection> connectionProvider) throws SQLException {
    	Transaction t = session.beginTransaction();

    	session.createQuery("DELETE FROM Person").executeUpdate();
    	session.connection().createStatement().executeUpdate("DELETE FROM " + SikredeRecordSpecs.ENTRY_RECORD_SPEC.getTable());
    	session.connection().createStatement().executeUpdate("DELETE FROM Yderregister_v3");
    	session.connection().createStatement().executeUpdate("DELETE FROM AssignedDoctor");
    	session.connection().createStatement().executeUpdate("DELETE FROM v2_ForaeldreMyndighedRelation");
    	session.connection().createStatement().executeUpdate("DELETE FROM v2_BarnRelation");
    	session.connection().createStatement().executeUpdate("DELETE FROM PersonKontakt");
    	session.connection().createStatement().executeUpdate("DELETE FROM HistoriskePersonnumre");
        session.connection().createStatement().executeUpdate("DELETE FROM PersonFamilie");
        session.connection().createStatement().executeUpdate("DELETE FROM v2_UmyndiggoerelseVaergeRelation");

    	for (Person person : persons) {
    	    session.persist(person);
            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
            String tilflytningsDato = dateFormat.format(person.getStatusDato());
            String sql = "INSERT INTO PersonKontakt (CPR, TilflytningKommuneDato, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','2000-01-01 00:00:00',now(),now(),now())";
            session.connection().createStatement().executeUpdate(sql);
            sql = "INSERT INTO HistoriskePersonnumre (CPR, Id, HistoriskCPR, CreatedDate, ModifiedDate, ValidFrom, ValidTo) VALUES ('"+person.getCpr()+"','','0000000000',now(),now(),'2000-01-01 00:00:00',now())";
            session.connection().createStatement().executeUpdate(sql);
            sql = "INSERT INTO HistoriskePersonnumre (CPR, Id, HistoriskCPR, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','','"+PRIOR_CPR+"',now(),now(),now())";
            session.connection().createStatement().executeUpdate(sql);
            if (familieMap.containsKey(person.getCpr())) {
                Person spouse = findPerson(familieMap.get(person.getCpr()));
                if (person.getKoen().equals(spouse.getKoen())) {
                    sql = "INSERT INTO PersonFamilie (CPR, Civilstand, AegtefaelleCPR, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','P','"+familieMap.get(person.getCpr())+"',now(),now(),now())";
                } else {
                    sql = "INSERT INTO PersonFamilie (CPR, Civilstand, AegtefaelleCPR, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','G','"+familieMap.get(person.getCpr())+"',now(),now(),now())";
                }
                session.connection().createStatement().executeUpdate(sql);
            }
            if (seperatedMap.containsKey(person.getCpr())) {
                sql = "INSERT INTO PersonFamilie (CPR, Civilstand, AegtefaelleCPR, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','F','"+seperatedMap.get(person.getCpr())+"',now(),now(),now())";
                session.connection().createStatement().executeUpdate(sql);
            }
            if (unmarriedList.contains(person.getCpr())) {
                sql = "INSERT INTO PersonFamilie (CPR, Civilstand, CreatedDate, ModifiedDate, ValidFrom) VALUES ('"+person.getCpr()+"','U',now(),now(),now())";
                session.connection().createStatement().executeUpdate(sql);
            }
            if (vaergeMap.containsKey(person.getCpr())) {
                sql = "INSERT INTO v2_UmyndiggoerelseVaergeRelation (Id, CPR, TypeKode, TypeTekst, RelationCpr, CreatedDate, ModifiedDate, ValidFrom, ValidTo) VALUES ('"+person.getCpr()+"-1','"+person.getCpr()+"','1','Værges CPR findes','"+vaergeMap.get(person.getCpr())+"',now(),now(),now(),'2100-12-12')";
                session.connection().createStatement().executeUpdate(sql);
            }
    	}

    	t.commit();

    	DateTime dt = new DateTime();
    	dt = dt.minusDays(1);

    	for (Record sikredeRecord: sikredeRecords) {
    		//RecordPersister should be injected
    		RecordPersister recordPersister = new RecordPersister(connectionProvider, dt.toInstant());
    		recordPersister.persist(sikredeRecord, SikredeRecordSpecs.ENTRY_RECORD_SPEC);
    	}

    	for (Record yderRecord: yderRecords) {
    		//RecordPersister should be injected
    		RecordPersister recordPersister = new RecordPersister(connectionProvider, dt.toInstant());
    		recordPersister.persist(yderRecord, YderregisterRecordSpecs.YDER_V3_RECORD_TYPE);
    	}

    	if (fmrRecords != null) {
    		for (Record fmrRecord: fmrRecords) {
    			RecordPersister recordPersister = new RecordPersister(connectionProvider, dt.toInstant());
    			recordPersister.persist(fmrRecord, CprRecordSpecs.FORAELDRE_MYNDIGHED_RELATION_RECORD_TYPE);
    		}
    	}

    	if (brRecords != null) {
    		for (Record brRecord: brRecords) {
    			RecordPersister recordPersister = new RecordPersister(connectionProvider, dt.toInstant());
    			recordPersister.persist(brRecord, CprRecordSpecs.BARN_RELATION_RECORD_TYPE);
    		}
    	}
    }

    public void clear() {
        persons.clear();
    	sikredeRecords.clear();
    	yderRecords.clear();
    	fmrRecords.clear();
    	brRecords.clear();
    }

    public static void assertDates(DateTime date1, Date date2) {
        assertDates(date1.toDate(), date2);;
    }

    public static void assertDates(Date date1, Date date2) {
        assertNotNull(date1);
        assertNotNull(date2);
        if (date2 instanceof java.sql.Timestamp) {
            date2 = new Date(date2.getTime());
        }
        assertEquals(date1.toString(),date2.toString());
    }

    public static void extendSikredeRecord(Record sikredeRecord) {
        sikredeRecord.put("SYdernrGl", "4321");
        sikredeRecord.put("SSKGyldigFra", "2000-01-01");
        sikredeRecord.put("SSKGyldigTil", "2100-01-01");
        sikredeRecord.put("SBSStatsborgerskabKode", "GB");
        sikredeRecord.put("SBSStatsborgerskab", "England");
        sikredeRecord.put("SSKBopelsLAndKode", "GE");
        sikredeRecord.put("SSKBopelsLand", "Tyskland");
        sikredeRecord.put("SSLSocSikretLandKode", "DK");
        sikredeRecord.put("SSLSocSikretLand", "Danmark");
    }
}
